import pytest
import warnings
import numpy as np
from numpy.testing import assert_allclose

from yaiv.defaults.config import ureg
from yaiv import utils as ut
from yaiv import spectrum, grep


def test_check_unit_consistency_success():
    ut._check_unit_consistency([1, 2.0, None])  # all unitless
    ut._check_unit_consistency(
        [1.0 * ureg.meter, 2.0 * ureg.second, None]
    )  # all unitful


def test_check_unit_consistency_failure(capsys):
    with pytest.raises(TypeError):
        ut._check_unit_consistency(
            [1.0 * ureg.meter, 2.0, None],
            names=["length", "value", "optional"],
        )
    out = capsys.readouterr().out
    assert "Units check failed for:" in out
    assert "Units status:" in out


def test_invQ():
    # Inverse should have 1/unit
    A = np.eye(2) * (2.0 * ureg.meter)
    inv = ut.invQ(A)
    assert isinstance(inv, ureg.Quantity)
    assert_allclose(inv.magnitude, 0.5 * np.eye(2))
    # Units: 1/meter
    assert inv.check(1 / ureg.meter)


def test_reciprocal_basis():
    # Lattice basis as identity (rows), expect reciprocal = 2π I
    a = 2 * np.eye(3) * ureg.meter
    K = ut.reciprocal_basis(a)
    assert isinstance(K, ureg.Quantity)
    assert_allclose(K.magnitude, 0.5 * np.eye(3))
    # Units: 2π / meter
    assert K.check(ureg._2pi / ureg.meter)


def test_cartesian_crystal_conversion():
    # Cubic 1 Å box
    a0 = 3.0 * ureg.angstrom
    basis = np.eye(3) * a0
    r_cart = np.array([0.3, 0.4, 0.5]) * ureg.angstrom

    r_cryst = ut.cartesian2cryst(r_cart, basis)
    # Should be in crystal units
    assert isinstance(r_cryst, ureg.Quantity)
    assert r_cryst.check(ureg.crystal)

    r_back = ut.cryst2cartesian(r_cryst, basis)
    assert isinstance(r_back, ureg.Quantity)
    assert r_back.check(ureg.angstrom)
    assert_allclose(
        r_back.to(ureg.angstrom).magnitude,
        r_cart.to(ureg.angstrom).magnitude,
        atol=1e-12,
    )


def test_cartesian2cryst_unit_incompatibility_raises():
    basis = np.eye(3) * (1.0 * ureg.meter)
    coord = np.array([1.0, 0.0, 0.0]) * ureg.meter
    with pytest.raises(TypeError):
        _ = ut.cartesian2cryst(coord.magnitude, basis)
    with pytest.raises(TypeError):
        _ = ut.cartesian2cryst(coord, basis.magnitude)
    with pytest.raises(TypeError):
        _ = ut.cartesian2cryst(coord.magnitude, basis.magnitude / ureg.meter)


def test_cryst2cartesian_unit_incompatibility_raises():
    basis = np.eye(3) * (1.0 * ureg.angstrom)
    coord = np.array([1.0, 0.0, 0.0])
    with pytest.raises(TypeError):
        _ = ut.cryst2cartesian(coord * ureg.meter, basis)
    with pytest.raises(TypeError):
        _ = ut.cryst2cartesian(coord, basis)


def test_voigt_cartesian_conversion():
    T = (
        np.array(
            [
                [1.0, 0.2, -0.3],
                [0.2, 2.0, 0.4],
                [-0.3, 0.4, -1.0],
            ]
        )
        * ureg.gigapascal
    )

    v = ut.cartesian2voigt(T)
    assert isinstance(v, ureg.Quantity)
    assert v.check(ureg.gigapascal)
    assert v.shape == (6,)

    T2 = ut.voigt2cartesian(v)
    assert isinstance(T2, ureg.Quantity)
    assert T2.check(ureg.gigapascal)
    assert_allclose(T2.magnitude, T.magnitude)


@pytest.mark.parametrize(
    "grid,expected_shape",
    [([1], (1, 1)), ([2], (2, 1)), ([2, 3], (6, 2)), ([2, 3, 4], (24, 3))],
)
def test_grid_generator_nonperiodic(grid, expected_shape):
    coords = ut.grid_generator(grid, periodic=False)
    assert coords.shape == expected_shape
    # Nonperiodic default spans [-1, 1], endpoints included (unless g==1)
    assert np.all(coords <= 1.0 + 1e-12)
    assert np.all(coords >= -1.0 - 1e-12)


def test_grid_generator_periodic():
    # periodic -> values in (-0.5, 0.5], no duplicates at borders
    coords = ut.grid_generator([4, 2], periodic=True)
    assert coords.shape == (8, 2)
    assert np.all(coords[:, 0] > -0.5 - 1e-12) and np.all(coords[:, 0] <= 0.5 + 1e-12)
    assert np.all(coords[:, 1] > -0.5 - 1e-12) and np.all(coords[:, 1] <= 0.5 + 1e-12)


def test_methpax_kernel_integrates_to_A():
    # Order 0 should integrate to A. Numerically approximate over a wide range.
    A = 1.7
    s = 0.2  # smearing
    xs = np.linspace(-5 * s, 5 * s, 2001)
    # order 0
    ys = np.array(
        [ut.methpax_kernel(x, mean=0.0, smearing=s, order=0, A=A) for x in xs]
    )
    integral = np.trapezoid(ys, xs)
    assert_allclose(
        integral, A, rtol=5e-3, atol=5e-3
    )  # loose tolerance due to finite range
    # order 1
    ys = np.array(
        [ut.methpax_kernel(x, mean=0.0, smearing=s, order=0, A=A) for x in xs]
    )
    integral = np.trapezoid(ys, xs)
    assert_allclose(
        integral, A, rtol=5e-3, atol=5e-3
    )  # loose tolerance due to finite range


def test_fermidirac_kernel_integrates_to_one():
    # Order 0 should integrate to A. Numerically approximate over a wide range.
    s = 0.2  # smearing
    xs = np.linspace(-10 * s, 10 * s, 2001)
    # order 0
    ys = ut.fermidirac_kernel(xs, mean=0.0, smearing=s)
    integral = np.trapezoid(ys, xs)
    assert_allclose(
        integral, 1, rtol=5e-3, atol=5e-3
    )  # loose tolerance due to finite range


def test_analyze_distribution_gaussian():
    mu = 0.3
    sigma = 0.1
    A = 1
    X = np.linspace(mu - 6 * sigma, mu + 6 * sigma, 2001)
    #    Y = (1.0 / (sigma * np.sqrt(2 * np.pi))) * np.exp(-0.5 * ((X - mu) / sigma) ** 2)
    Y = ut._normal_dist(X, mean=mu, sd=sigma, A=A)

    stats = ut.analyze_distribution(X, Y)
    assert_allclose(stats.norm, A, rtol=1e-3, atol=1e-3)
    assert_allclose(stats.mean, mu, rtol=1e-3, atol=1e-3)
    assert_allclose(stats.std, sigma, rtol=1e-3, atol=1e-3)
    # Gaussian skewness ~ 0, kurtosis ~ 3
    assert abs(stats.skewness) < 1e-2
    assert_allclose(stats.kurtosis, 3.0, rtol=1e-2, atol=1e-2)


def test_analyze_distribution_zero_norm_raise():
    X = np.linspace(-2, 2, 201)
    Y = X
    with pytest.raises(ValueError):
        ut.analyze_distribution(X, Y)


def test_kernel_density_normalization_gaussian():
    """
    For Gaussian kernel (order=0), the integral over the density should equal
    sum_k w_k * sum_b values[k,b]. With values=1 and weights summing to 1,
    this equals the number of bands.
    """
    nk, nb = 4, 3
    # Simple energies spread around 0
    x = np.array(
        [
            [-0.20, 0.00, 0.15],
            [-0.10, 0.05, 0.25],
            [0.00, 0.10, 0.30],
            [0.05, 0.20, 0.40],
        ]
    )
    assert x.shape == (nk, nb)
    values = np.ones_like(x)
    weights = np.ones(nk) / nk  # sum(weights) = 1

    # Choose a window around the energies, and small sigma for reasonable resolution
    center = 0.0
    x_window = 1.0
    sigma = 0.05
    steps = 2000

    out = ut.kernel_density_on_grid(
        x=x,
        values=values,
        weights=weights,
        center=center,
        x_window=x_window,
        sigma=sigma,
        steps=steps,
        order=0,  # Gaussian
        cutoff_sigmas=5.0,
    )
    # Integral of density over grid should be nb (sum_b values) since sum_k w_k = 1
    integral = np.trapezoid(out.density, out.grid)
    assert_allclose(integral, nb, rtol=5e-3, atol=5e-3)


def test_kernel_density_units():
    """
    With unitful inputs: grid should have x units; density should have units values/x.
    """
    nk, nb = 2, 2
    x = np.array([[0.0, 0.1], [0.2, 0.3]]) * ureg.eV
    values = np.ones_like(x)  # dimensionless "states"
    weights = np.array([0.5, 0.5])

    out = ut.kernel_density_on_grid(
        x=x,
        values=values,
        weights=weights,
        center=0.0 * ureg.eV,
        x_window=0.5 * ureg.eV,
        sigma=0.05 * ureg.eV,
        steps=1000,
        order=0,
    )
    # Grid units should be eV
    assert hasattr(out.grid, "units")
    assert out.grid.check(ureg.eV)
    # Density units should be 1/eV
    assert hasattr(out.density, "units")
    assert out.density.check(1 / ureg.eV)


def test_kernel_density_shape_errors():
    """
    Mismatched shapes should raise ValueError.
    """
    nk, nb = 3, 2
    x = np.random.rand(nk, nb)
    values_bad = np.random.rand(nk, nb + 1)  # wrong shape
    weights_bad = np.ones(nk + 1) / (nk + 1)  # wrong length

    with pytest.raises(ValueError):
        ut.kernel_density_on_grid(x=x, values=values_bad)

    with pytest.raises(ValueError):
        ut.kernel_density_on_grid(x=x, values=np.ones_like(x), weights=weights_bad)


def test_kernel_density_mp_order_1_normalization():
    """
    For MP kernel with order=1, the integral should still approximate sum_k w_k * sum_b values,
    provided the window and cutoff resolve the kernel tails well.
    """
    nk, nb = 3, 2
    x = np.array(
        [
            [-0.15, 0.05],
            [0.00, 0.10],
            [0.20, 0.30],
        ]
    )
    values = np.ones_like(x)
    weights = np.ones(nk) / nk

    out = ut.kernel_density_on_grid(
        x=x,
        values=values,
        weights=weights,
        center=0.0,
        x_window=1.0,
        sigma=0.06,
        steps=2000,
        order=1,  # Methfessel–Paxton order 1
        cutoff_sigmas=5.0,
    )
    integral = np.trapezoid(out.density, out.grid)
    assert_allclose(integral, nb, rtol=1e-2, atol=1e-2)


def test_kernel_density_window_and_steps_defaults():
    """
    If x_window and steps are None, the function should still produce a sensible grid and density.
    """
    nk, nb = 2, 2
    x = np.array([[0.0, 0.1], [0.2, 0.3]])
    out = ut.kernel_density_on_grid(
        x=x, values=None, weights=None, sigma=None, steps=None, order=0
    )
    # Basic sanity checks
    assert isinstance(out.grid, np.ndarray) or hasattr(out.grid, "magnitude")
    assert isinstance(out.density, np.ndarray) or hasattr(out.density, "magnitude")
    # density and grid lengths must match
    if isinstance(out.grid, np.ndarray):
        assert out.grid.shape == out.density.shape
    else:
        assert out.grid.magnitude.shape == out.density.magnitude.shape


def test_kernel_regression_units_and_constant_recovery():
    """
    Output units must match `values` units, and a constant `values` should be recovered.
    """
    nk, nb = 3, 2
    # x with units (e.g., eV), and constant values with another unit (e.g., ampere)
    x = np.array([[-0.3, 0.0], [0.1, 0.2], [0.4, 0.5]]) * ureg.eV
    values = np.full_like(x, 5.0) * ureg.ampere
    weights = np.ones(nk) / nk

    # Build regression callable
    f = ut.kernel_regression(
        x=x,
        values=values,
        weights=weights,
        default_sigma=0.05 * ureg.eV,
        default_cutoff_sigmas=5.0,
        order=0,
    )

    # Evaluate on a grid in the same x units
    X = np.linspace(-0.4, 0.6, 201) * ureg.eV
    y = f(X)

    # Units must match values units
    assert hasattr(y, "units")
    assert y.check(ureg.ampere)

    # For constant values, the regression output should be approximately the constant
    assert_allclose(y.magnitude, 5.0, rtol=5e-2, atol=5e-2)


def test_kernel_regression_sigma_override_and_scalar_eval():
    """
    The returned callable should accept scalar/array inputs and allow overriding sigma at call time.
    """
    nk, nb = 2, 2
    x = np.array([[0.0, 0.1], [0.3, 0.4]])  # unitless for simplicity
    # Let values be constant so the target is easy to check
    values = np.ones_like(x) * 2.0
    weights = np.ones(nk) / nk

    # Build callable with a default sigma
    f = ut.kernel_regression(
        x=x,
        values=values,
        weights=weights,
        default_sigma=0.05,
        default_cutoff_sigmas=5.0,
        order=0,
    )

    # Scalar evaluation
    y0 = f(0.2)
    # Array evaluation with an override sigma (narrower kernel)
    X = np.array([0.05, 0.2, 0.35])
    y1 = f(X, sigma=0.02)

    # Outputs are numeric, shapes as expected
    assert np.all(y0 != y1)
    assert np.isscalar(y0) or getattr(y0, "shape", ()) == ()
    assert isinstance(y1, np.ndarray)

    # Since values are constant (2.0), regression output should be ~2.0
    # (tolerate some smoothing variation)
    assert_allclose(y0, 2.0, rtol=5e-2, atol=5e-2)
    assert_allclose(y1, 2.0, rtol=5e-2, atol=5e-2)


def test_expand_zone_border_shapes_and_units():
    q = np.array([[0.5, 0.0, 0.0], [0, 0, 0]]) * ureg.meter
    with pytest.raises(TypeError):
        expanded = ut._expand_zone_border(q)
    q = np.array([0.5, 0.0, 0.0])
    expanded = ut._expand_zone_border(q)
    assert expanded.shape == (27, 3)
    q = np.array([[0.5, 0.0, 0.0], [0, 0, 0]])
    expanded = ut._expand_zone_border(q)
    assert expanded.shape == (2 * 27, 3)
    # With crystal units preserved
    q_u = q * ureg.crystal
    expanded_u = ut._expand_zone_border(q_u)
    assert isinstance(expanded_u, ureg.Quantity)
    assert expanded_u.check(ureg.crystal)
    assert_allclose(expanded_u.magnitude, expanded)


def test_amplitude2order_parameter():
    # Two atoms, simple displacements
    amps = np.array([0.05, 0.10]) * ureg.angstrom  # length units
    masses = np.array([1.0, 4.0]) * ureg.kilogram  # mass units
    disp0 = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0]])  # norms: 1 and 2
    disp1 = np.array([[0.0, 1.0, 0.0], [0.0, 0.0, 3.0]])  # norms: 1 and 3
    # NN_i = sum_j M_j * ||ε_ij||^2
    # For disp0: 1*1^2 + 4*2^2 = 1 + 16 = 17
    # For disp1: 1*1^2 + 4*3^2 = 1 + 36 = 37
    expected = np.array(
        [amps.magnitude[0] * np.sqrt(17), amps.magnitude[1] * np.sqrt(37)]
    )

    q = ut.amplitude2order_parameter(amps, masses, [disp0, disp1])
    assert isinstance(q, ureg.Quantity)
    # Units: length * sqrt(mass)
    assert q.check("angstrom * kilogram^0.5")
    assert_allclose(q.magnitude, expected)


def test_cumulative_integral_input_validation():
    with pytest.raises(ValueError):
        ut.cumulative_integral(np.array([[0, 1]]), np.array([0, 1]))  # x not 1D
    with pytest.raises(ValueError):
        ut.cumulative_integral(np.array([0, 1]), np.array([0, 1, 2]))  # shape mismatch
    with pytest.raises(ValueError):
        ut.cumulative_integral(
            np.array([0, 0, 1]), np.array([0, 1, 2])
        )  # not strictly increasing


def test_cumulative_integral_linear_function_and_units():
    # y = x => integral = 0.5 x^2
    x = np.linspace(0.0, 2.0, 201)
    y = x.copy()
    I = ut.cumulative_integral(x, y)
    assert_allclose(I, 0.5 * x**2, rtol=1e-6, atol=1e-6)

    # With units: x in m, y in N => integral has units N*m
    x_u = x * ureg.meter
    y_u = y * ureg.newton
    Iu = ut.cumulative_integral(x_u, y_u)
    assert isinstance(Iu, ureg.Quantity)
    assert Iu.check(ureg.newton * ureg.meter)
    assert_allclose(Iu.magnitude, I, rtol=1e-6, atol=1e-6)


def test_point_to_segment_distance():
    # Point on the segment
    A = np.array([0.0, 0.0, 0.0])
    B = np.array([1.0, 0.0, 0.0])
    P_on = np.array([0.3, 0.0, 0.0])
    assert_allclose(ut._point_to_segment_distance(P_on, A, B), 0.0)

    # Point above the middle
    P_above = np.array([0.5, 2.0, 0.0])
    assert_allclose(ut._point_to_segment_distance(P_above, A, B), 2.0)

    # Point nearest to endpoint A
    P_near_A = np.array([-1.0, 1.0, 0.0])
    # The closest point on the segment is A (0,0,0), distance sqrt(2)
    assert_allclose(
        ut._point_to_segment_distance(P_near_A, A, B), np.sqrt(2.0), atol=1e-12
    )


def test_symmetry_orbit_kpoints_matches_2x2x2_grid(data_dir, require):
    """
    For a QE XML with a 2x2x2 Monkhorst-Pack grid, the symmetry orbit
    of IBZ k-points (modulo G) should match the full 2x2x2 grid generated
    by ut.grid_generator (up to ordering).
    """

    fname = data_dir / "qe/results_scf/scf.xml"
    require(fname, f"Missing test data: {fname}")

    # Read symmetries and k-points from file
    syms = grep.symmetries(str(fname))
    data = spectrum.ElectronBands(str(fname))
    # Quantity in _2pi/crystal (expected)
    k_ibz = ut.cartesian2cryst(data.kpoints / data.alat, data.k_lattice)

    # Compute symmetry orbit (modulo G)
    out = ut.symmetry_orbit_kpoints(k_ibz, syms, tol=1e-12, mod_G=True)

    # Build expected 2x2x2 grid in crystal units
    # periodic=True means values in (-0.5, 0.5] for a 2x2x2 mesh
    grid = ut.grid_generator([2, 2, 2], periodic=True)  # ndarray or Quantity

    # Prepare actual orbit set for comparison (float crystal units)
    orbit = out.kpoints.magnitude

    # Compare sets ignoring order: sort rows and compare shape and values
    assert grid.shape == orbit.shape

    # For each expected point, ensure there is an orbit point equivalent mod 1
    def is_close_mod1(a, b, tol=1e-12):
        """
        Check if vectors a and b are equal modulo integers (component-wise), i.e.,
        a - b ≡ 0 (mod 1). Assumes a, b are 1D arrays of length 3.
        """
        d = a - b
        d = d - np.round(d)
        return np.all(np.abs(d) <= tol)

    for kg in grid:
        assert any(
            is_close_mod1(kg, ko, tol=1e-12) for ko in orbit
        ), f"Missing k-point mod 1: {ke}"

    # mod_G requires crystal units; verify unit on output
    assert hasattr(out.kpoints, "units")
    assert out.kpoints.check(ureg("_2pi/crystal"))


def test_find_little_group_silicon(data_dir, require):
    fname = data_dir / "qe/results_scf/scf.xml"
    require(fname, f"Missing test data: {fname}")

    # Read symmetries and k-points from file
    syms = grep.symmetries(str(fname))
    data = spectrum.ElectronBands(str(fname))
    # Quantity in _2pi/crystal (expected)
    k_ibz = ut.cartesian2cryst(data.kpoints / data.alat, data.k_lattice)

    # Compute little_group or original points and points in the orbits
    origin_lg = ut.find_little_group(k_ibz, syms, tol=1e-12)
    orbit = ut.symmetry_orbit_kpoints(k_ibz, syms, tol=1e-12, mod_G=True)
    orbit_lg = ut.find_little_group(orbit.kpoints, syms, tol=1e-12)

    # Check that all little groups of points in the star have the same number of elements
    for i, k in enumerate(orbit.kpoints):
        assert len(orbit_lg[i]) == len(origin_lg[orbit.origin[i]])


def _avg_delta_k(lattice, kgrid):
    # Matches the average Δk definition used in verbose mode
    if isinstance(lattice, ureg.Quantity):
        Kvol = (2 * np.pi) ** 3 / np.linalg.det(lattice.magnitude)
        dk = (Kvol / np.prod(kgrid)) ** (1 / 3) * (1 / lattice.units)
        return dk
    else:
        Kvol = (2 * np.pi) ** 3 / np.linalg.det(lattice)
        return (Kvol / np.prod(kgrid)) ** (1 / 3)


def test_auto_kgrid_spacing_and_parity():
    # Cubic 5 Å cell, target Δk = 0.1 Å^-1
    a = 5.0 * ureg.angstrom
    lattice = np.eye(3) * a
    dk_target = 0.1 / ureg.angstrom

    # Enforce odd in x,z and even in y
    kgrid = ut.auto_kgrid(
        lattice=lattice,
        delta_k=dk_target,
        force_odd=[True, False, True],
        force_even=[False, True, False],
    )

    # Basic sanity
    assert isinstance(kgrid, list) and len(kgrid) == 3
    assert all(isinstance(ni, int) and ni >= 1 for ni in kgrid)

    # Parity checks
    assert kgrid[0] % 2 == 1  # odd
    assert kgrid[1] % 2 == 0  # even
    assert kgrid[2] % 2 == 1  # odd

    # Achieved average Δk close to target (allow ~25% due to integer rounding/parity)
    dk_eff = _avg_delta_k(lattice, kgrid)
    assert dk_eff.check(ureg.angstrom**-1)
    rel_err = (
        abs(dk_eff.to(dk_target.units).magnitude - dk_target.magnitude)
        / dk_target.magnitude
    )
    assert rel_err < 0.25


def test_eigen_projection():
    # BASIC projection
    # Setup initial and final spectra
    spectra_in = np.array([[1, 0], [0, 1]])
    spectra_out = np.array([[1, 0], [0, 1]])

    # Test without eigenvalues_out, expecting identity projections
    proj = ut.eigen_projection(spectra_in, spectra_out)
    expected_proj = np.array([[1, 0], [0, 1]], dtype=complex)
    np.testing.assert_array_almost_equal(proj, expected_proj)

    # BASIC projection with degenerate eigenvalues
    spectra_in = np.array([[1, 0], [0, 1]])
    spectra_out = np.array([[1, 0], [0, 1]])
    eigenvalues_out = np.array([1.0, 1.0])  # Degenerate eigenvalues

    proj = ut.eigen_projection(spectra_in, spectra_out, eigenvalues_out)
    expected_proj = np.array([[1, 0], [1, 0]], dtype=float)
    np.testing.assert_array_almost_equal(proj, expected_proj)

    # Setup spectra with unitful eigenvalues using Pint
    spectra_in = np.array([[1, 0], [0, 1]])
    spectra_out = np.array([[1, 0], [0, 1]])
    eigenvalues_out = np.array([1.0, 1.0]) * ureg.hertz

    proj = ut.eigen_projection(spectra_in, spectra_out, eigenvalues_out)
    expected_proj = np.array([[1, 0], [1, 0]], dtype=float)
    np.testing.assert_array_almost_equal(proj, expected_proj)

    # Setup spectra that violate norm condition
    spectra_in = np.array([[1, 1] / np.sqrt(2), [0, 0]])
    spectra_out = np.array([[1, 0], [0, 1]])

    # Check that a warning is raised
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        ut.eigen_projection(spectra_in, spectra_out)
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "Projections norm is" in str(w[-1].message)



"""
YAIV | yaiv.cell
================

This module defines core functions and a container class for crystal structures
used in symmetry analysis, format conversion, and structural manipulation.

It provides a `Cell` class that wraps an ASE Atoms object along with its spglib-compatible
representation. The `Cell` object allows for easy integration with spglib and includes
utility methods to extract and report symmetry information, Wyckoff positions, and
symmetry operations in symbolic form.

The module also includes conversion tools between ASE and spglib formats,
as well as symmetry naming utilities.

Classes
-------
Cell
    Wrapper around ASE Atoms + spglib tuple (lattice, positions, numbers).
    Provides:
    - from_file(path): Read a file and return Cell.
    - from_spglib_tuple(tup): Construct Cell from spglib-format tuple.
    - get_sym_info(symprec=...): Print symbolic symmetry operations and Wyckoff info.
    - get_wyckoff_positions(symprec=...): Group atoms by Wyckoff positions.
    - get_supercell(supercell): Return a repeated Cell object.
    - write_espresso_in(...): Write Quantum ESPRESSO input file.
    - print(...): Write the crystal structure in a human‑readable text format.
    - view(...): Visualize the atomic structure in 3D.

Functions
---------
ase2spglib(crystal_ase)
    Convert an ASE Atoms object to a spglib tuple (lattice, positions, numbers).

spglib2ase(spglib_crystal)
    Convert a spglib-format tuple back into an ASE Atoms object.

read_spg(file)
    Read a structure file and return its spglib-compatible tuple.

Private Utilities
-----------------
_rot_name(rot, lattice)
    Identify the symbolic name and axis of a rotation matrix (e.g. 'C3', 'm') using its eigenstructure.

Examples
--------
>>> from yaiv.cell import Cell
>>> cell = Cell.from_file("data/POSCAR")
>>> cell.get_sym_info()
SpaceGroup = Fd-3m (227)
...
>>> cell.get_wyckoff_positions()
>>> cell.wyckoff.labels
['a', 'b']
>>> cell.wyckoff.symbols
['Si', 'Si']

See Also
--------
yaiv.defaults : Configuration and default precision values
yaiv.utils    : Utility functions for basis and vector transformations
"""

from types import SimpleNamespace

import numpy as np

# from ase.io import read, write
from ase import Atoms
import nglview as nv

from yaiv import utils as ut
from yaiv.defaults.config import defaults

__all__ = ["read", "write", "read_spg", "ase2spglib", "spglib2ase", "Cell"]


class Cell:
    """
    A wrapper that stores both an ASE Atoms object and the corresponding
    spglib-format tuple (lattice, positions, numbers).

    This class allows use in spglib (via tuple interface) and in ASE (via .atoms).
    Moreover, it adds some extra functionality with custom methods.

    Attributes
    ----------
    atoms : ase.Atoms
        Full ASE Atoms object with chemical info.
    spglib : tuple
        Tuple (lattice, positions, numbers) derived from the Atoms object,
        used for spglib symmetry operations.

    Methods
    -------
    from_file(...)
        Read a structure file using ASE and return a Cell instance.
    from_spglib_tuple(...)
        Initialize from a (lattice, positions, numbers) spglib tuple.
    get_sym_info(...)
        Print a detailed report of symmetry information for the crystal structure.
    get_wyckoff_positions(...)
        Analyze the structure and store information about independent Wyckoff positions.
    get_supercell(...)
        Construct a supercell by repeating the current unit cell along each lattice direction.
    write_espresso_in(...)
        Write Quantum ESPRESSO input file using either default parameters or a template.
    print(...)
        Write the crystal structure in a human‑readable text format.
    view(...)
        Visualizes the atomic structure in 3D.
    """

    def __init__(
        self,
        lattice: np.ndarray = None,
        positions: np.ndarray = None,
        numbers: np.ndarray = None,
        atoms: Atoms = None,
    ):
        """
        Initialize a Cell object.

        Parameters
        ----------
        lattice : array-like, optional
            3x3 lattice matrix.
        positions : array-like, optional
            Nx3 array of fractional coordinates.
        numbers : array-like, optional
            Length-N array of atomic numbers.
        atoms : Atoms, optional
            Alternative way to initialize with an Atoms object.

        Raises
        ------
        ValueError
            If neither individual arguments nor `ase.Atoms` object is provided correctly.
        """
        if atoms is not None:
            if not isinstance(atoms, Atoms):
                raise ValueError("`atoms` must be an ase.Atoms object.")
            self.atoms = atoms
            self.spglib = ase2spglib(atoms)
        elif lattice is None or positions is None or numbers is None:
            raise ValueError(
                "Must provide either individual components or an `ase.Atoms` object."
            )
        else:
            self.spglib = (np.array(lattice), np.array(positions), np.array(numbers))
            self.atoms = spglib2ase(self.spglib)

    @classmethod
    def from_file(cls, file: str):
        """
        Read a structure file using ASE and return a Cell instance.

        Parameters
        ----------
        file : str
            Path to structure file (e.g. CIF, POSCAR).

        Returns
        -------
        Cell
            A new Cell instance with Atoms and spglib data.
        """
        from ase.io import read

        atoms = read(file)
        return cls(atoms=atoms)

    @classmethod
    def from_spglib_tuple(cls, tup):
        """
        Initialize from a (lattice, positions, numbers) spglib tuple.

        Parameters
        ----------
        tup : tuple
            A 3-tuple (lattice, positions, numbers)

        Returns
        -------
        Cell
            A new Cell instance with Atoms and spglib data.
        """
        lattice, positions, numbers = map(np.array, tup)
        return cls(lattice, positions, numbers)

    def __iter__(self):
        return iter(self.spglib)

    def __getitem__(self, key):
        return self.spglib[key]

    def __len__(self):
        return len(self.spglib)

    def __repr__(self):
        return f"<Cell with {len(self.spglib[2])} atoms>\n\n{self.spglib}"
        return f"\n{self.spglib}"

    def get_sym_info(self, symprec: float = defaults.symprec):
        """
        Print a detailed report of symmetry information for the crystal structure.

        This includes:
        - The chemical formula.
        - The space group name and number.
        - Wyckoff positions and equivalent atoms.
        - Site symmetry symbols.
        - A list of all symmetry operations (rotations + translations), with symbolic names (E, Cn, m, I, Sn).

        Parameters
        ----------
        symprec : float, optional
            Symmetry tolerance used by spglib to determine symmetry operations.
            The default is 1e-5. This sets the precision for atomic position comparison.

        Notes
        -----
        This is a diagnostic utility that gives insight into the symmetry content of the crystal.
        It is not meant for structured programmatic use, but rather for human-readable output.
        """
        from spglib import get_symmetry_dataset

        atoms = spglib2ase(self.spglib)
        print(atoms.get_chemical_formula())
        dataset = get_symmetry_dataset(self.spglib, symprec=symprec)
        print("SpaceGroup =", dataset.international, "(" + str(dataset.number) + ")")
        print()
        print("ATOMS:")
        print(atoms.get_chemical_symbols())
        print("Wyckoffs:")
        print(dataset.wyckoffs)
        print("Equivalent positions:")
        print(dataset.equivalent_atoms)
        print("Site symmetry simbols:")
        print(dataset.site_symmetry_symbols)
        print()
        print("Symmetry Operations:")
        print()
        symmetry = [(r, t) for r, t in zip(dataset.rotations, dataset.translations)]
        for i in range(len(symmetry)):
            rot = symmetry[i][0]
            t = np.around(symmetry[i][1], decimals=3)
            sym = _rot_name(rot, self.spglib[0])
            print(f"{sym.label}", end="")
            if sym.direction_crystal is not None:
                dir_lat = np.round(sym.direction_crystal, 3)
                dir_cart = np.round(sym.direction_cartesian, 3)
                print(f" / [a,b,c] = {dir_lat} / [x,y,z] = {dir_cart}")
            else:
                print()
            # _rot_name(rot, self.spglib[0])
            print(f"{rot} + {t}")

    def get_wyckoff_positions(self, symprec: float = defaults.symprec):
        """
        Analyze the structure and store information about independent Wyckoff positions.

        This method identifies groups of symmetry-equivalent atoms using spglib
        and organizes them by Wyckoff label. It stores the results in the attribute
        `self.wyckoff`, which is a SimpleNamespace with the following fields:

        Attributes
        ----------
        self.wyckoff : SimpleNamespace
            A namespace with grouped symmetry information:
                - symbols : list of str
                    Chemical symbol for each independent Wyckoff site.
                - labels : list of str
                    Wyckoff letter (e.g. 'a', 'b', 'c', ...) corresponding to each independent site.
                - positions : list of ndarray, shape (N_i, 3)
                    List of arrays, each containing the fractional coordinates of atoms in a symmetry-equivalent group.
                - indices : list of list of int
                    Indices (in the full structure) of atoms belonging to each Wyckoff group.

        Parameters
        ----------
        symprec : float, optional
            Symmetry tolerance for spglib. Atoms closer than this value are considered equivalent.
        """
        from spglib import get_symmetry_dataset

        spglib_data = self.spglib
        atoms = spglib2ase(spglib_data)
        dataset = get_symmetry_dataset(spglib_data, symprec=symprec)

        wyckoff_letters = dataset.wyckoffs
        equivalent_atoms = dataset.equivalent_atoms
        symbols = atoms.get_chemical_symbols()

        unique_labels = []
        unique_symbols = []
        grouped_positions = []
        grouped_indices = []

        seen_equiv = []

        for i, eq in enumerate(equivalent_atoms):
            if eq not in seen_equiv:
                # New independent site
                seen_equiv.append(eq)
                unique_symbols.append(symbols[i])
                unique_labels.append(wyckoff_letters[i])
                grouped_positions.append([spglib_data[1][i]])
                grouped_indices.append([i])
            else:
                # Append to existing site
                index = seen_equiv.index(eq)
                grouped_positions[index].append(spglib_data[1][i])
                grouped_indices[index].append(i)

        # Convert positions to arrays
        grouped_positions = [np.array(group) for group in grouped_positions]

        self.wyckoff = SimpleNamespace(
            symbols=unique_symbols,
            labels=unique_labels,
            positions=grouped_positions,
            indices=grouped_indices,
        )

    def get_supercell(self, supercell: list[int] = [1, 1, 1]) -> "Cell":
        """
        Construct a supercell by repeating the current unit cell along each lattice direction.

        This method generates a new `Cell` object representing a supercell formed by tiling
        the original cell along the three lattice vectors.

        Parameters
        ----------
        supercell : list[int], optional
            A list of 3 integers specifying how many times to replicate the cell along
            each lattice direction. Default is [1, 1, 1] (no replication).

        Returns
        -------
        Cell
            A new `Cell` object representing the expanded supercell.
        """

        lattice = np.copy(self[0])  # Original lattice vectors (3x3)
        positions_list = []

        for i in range(supercell[0]):
            for j in range(supercell[1]):
                for k in range(supercell[2]):
                    displacement = np.array([i, j, k])
                    new_pos = self[1] + displacement  # Shift all atoms
                    positions_list.append(new_pos)

        # Stack all positions and flatten to a single (N_atoms × 3) array
        positions = np.vstack(positions_list)

        # Repeat the atomic elements accordingly
        elements = np.tile(self[2], np.prod(supercell))

        # Expand the lattice
        for i in range(3):
            lattice[i] *= supercell[i]
            positions[:, i] /= supercell[i]  # Normalize positions to new cell

        return Cell(lattice, positions, elements)

    def write_espresso_in(
        self,
        filename: str = "espresso.pwi",
        template: str = None,
        kgrid: tuple | list = None,
    ):
        """
        Write Quantum ESPRESSO input file using either default parameters or a template.

        If no template is provided, writes a new input using default settings stored
        in `yaiv.defaults.config.qe_defaults`. If a template is provided, it replaces the structural
        information (cell, atomic positions, nat) in the template with those from
        `self.atoms`.

        Parameters
        ----------
        filename : str
            Output input file name (e.g. 'espresso.pwi').
        template : str
            Optional template input file to use as a base. Only geometry-related fields
            (CELL_PARAMETERS, ATOMIC_POSITIONS, nat) are updated.
        kgrid : list, optional
            Desiered number of kgrid [N1,N2,N3]. Defaults to the template or `qe_defaults`.
        """
        import re, os
        from ase.io import write
        from yaiv.defaults.config import qe_defaults

        # Build dummy dictionary for pseudopotentials (needed from ASE 3.24)
        species = set(self.atoms.get_chemical_symbols())
        pseudopotentials = {}
        for s in species:
            pseudopotentials[s] = f"{s}.upf"
        print(pseudopotentials)

        # Pass a valid kgrid tuple.
        if isinstance(kgrid, list):
            kpts = kgrid = tuple(kgrid)
        elif kgrid is None:
            kpts = qe_defaults.kpts

        # Generate a basic template with ASE if not provided
        if template is None:
            write(
                filename,
                self.atoms,
                input_data=qe_defaults.input_data,
                kpts=kpts,
                format="espresso-in",
                pseudopotentials=pseudopotentials,
            )
            return

        # Write a temporary QE input from the structure to extract updated geometry
        print(pseudopotentials)
        write(
            ".tmp.pwi",
            self.atoms,
            format="espresso-in",
            pseudopotentials=pseudopotentials,
        )

        # Extract updated geometry info: CELL_PARAMETERS, ATOMIC_POSITIONS, nat
        basis = []
        pos = []
        cell_line = -4
        pos_line = -999999
        nat = 0
        with open(".tmp.pwi", "r") as lines:
            for n, line in enumerate(lines):
                if re.search(r"\bnat\b", line):
                    nat = int(line.split()[2])
                if re.search("CELL_PARAMETERS", line):
                    cell_line = n
                if re.search("ATOMIC_POSITIONS", line):
                    pos_line = n
                if n - cell_line in [1, 2, 3]:
                    basis.append(line)
                if n - pos_line in range(1, nat + 1):
                    pos.append(line)
        os.remove(".tmp.pwi")

        # Open template and inject updated structural info
        write_nat = True
        write_pos = True
        write_basis = True
        write_kpoints = False

        temp = open(template, "r")
        output = open(filename, "w")
        for line in temp:
            if re.search("ibrav", line):
                if "0" not in line:
                    raise ValueError("ERROR: Your template must have ibrav = 0.")
            elif re.search("pseudo_dir", line):
                line = "  pseudo_dir = '$PSEUDO_DIR',\n"
            elif re.search("outdir", line):
                line = "  outdir = './tmp',\n"
            elif re.search("nat*=", line) and write_nat == True:
                line = "  nat=" + str(nat) + ",\n"
                write_nat = False
            elif re.search("POSITIONS", line, re.IGNORECASE) and write_pos == True:
                line = "ATOMIC_POSITIONS {angstrom}\n"
                output.write(line)
                for line in pos:
                    output.write(line)
                write_pos = False
            elif re.search("POINTS", line, re.IGNORECASE):
                write_kpoints = True
            elif re.search("CELL", line, re.IGNORECASE):
                line = "CELL_PARAMETERS {angstrom}\n"
                output.write(line)
                for line in basis:
                    output.write(line)
                write_kpoints = False
            if write_pos == True:
                output.write(line)
            elif write_kpoints == True:
                output.write(line)
                if kgrid is not None:
                    output.write("  " + " ".join(map(str, (*kgrid, 0, 0, 0))) + "\n")
                    write_kpoints = False
        temp.close()
        output.close()

    def print(self, filename: str = None):
        """
        Write the crystal structure in a human‑readable text format.

        If `filename` is provided, the output is written to that file. If `filename`
        is None, the output is printed to standard output.

        Parameters
        ----------
        filename : str, optional
            Path to the file to write. If None, writes to stdout.
        """
        from sys import stdout

        # Gather data from ASE Atoms
        cell = np.asarray(self.atoms.get_cell())
        positions = np.asarray(self.atoms.get_scaled_positions())
        symbols = self.atoms.get_chemical_symbols()

        # Decide output stream
        out = stdout if filename is None else open(filename, "w")

        try:
            # Write CELL (Angstrom)
            np.savetxt(out, cell, fmt="%14.9f", header="CELL (Angstrom)", comments="")
            out.write("\n")

            # Write atomic positions in crystal coordinates
            out.write("Atomic Positions (crystal)\n")
            for s, (x, y, z) in zip(symbols, positions):
                out.write(f"{s:<2} {x:14.9f} {y:14.9f} {z:14.9f}\n")
        finally:
            if out is not stdout:
                out.close()

    def view(
        self,
        repeat: tuple[int] = (1, 1, 1),
        cell: bool = True,
        lattice: bool = True,
        perspective: bool = False,
        size: float = 1,
        radius: float = 0.1,
    ) -> nv.widget.NGLWidget:
        """
        Visualizes the atomic structure in 3D, with options for repeating, cell display, lattice vectors, and perspective adjustments.

        This method uses the NGL Viewer to render an interactive 3D model of the atomic structure, allowing users to customize the visualization
        by adjusting the repetition of the unit cell, displaying lattice information, choosing between orthographic and perspective views, and
        resizing the widget.

        Parameters
        ----------
        repeat : tuple[int], optional
            Defines the number of repetitions along each crystallographic axis (x, y, z) to be shown. Defaults to (1, 1, 1), showing only the
            original unit cell.
        cell : bool, optional
            Indicates whether the unit cell box should be displayed within the visualization. Defaults to True.
        lattice : bool, optional
            Specifies whether the lattice vectors should be visualized using arrows. Defaults to True.
        perspective : bool, optional
            If True, sets the camera to perspective mode for a more realistic view. If False, uses orthographic mode for diagram-like accuracy.
            Defaults to False.
        size : float, optional
            Specifies the size multiplier for the widget dimensions. Larger values increase the pixel dimensions of the visualization. Defaults to 1.
        radius : float, optional
            Radius (size) of balls and bonds.

        Returns
        -------
        nv.widget.NGLWidget
            An interactive widget showing the atomic structure, complete with customizable features for user analysis and exploration.
        """
        # Create the widget with optional repeated structure
        widget = nv.show_ase(self.atoms.repeat(repeat), default=False)

        # Adjust widget dimensions using the `size` parameter
        widget.layout.width = f"{int(size * 800)}px"
        widget.layout.height = f"{int(size * 500)}px"
        widget.handle_resize()  # Update to apply new sizes

        # Configure camera view settings
        if not perspective:
            widget.camera = "orthographic"  # Set diagram-like camera perspective
        if cell:
            widget.add_unitcell()  # Add lattice box visualization
        # Display atoms and bonds clearly
        widget.add_representation("ball+stick", radius=radius, aspectRatio=5)
        if lattice:
            # Calculate normalized lattice vectors for visual representation
            vectors = (
                np.asarray([vec / np.linalg.norm(vec) for vec in self.spglib[0]]) * 2
            )
            origin = [0, 0, 0]
            # Color coding and names for vectors (RGB)
            colors = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            names = ["a", "b", "c"]
            for i in range(3):
                # Use arrow drawing capability to show lattice directions
                widget.shape.add_arrow(origin, vectors[i], colors[i], 0.2, names[i])

        # Apply initial rotations and zoom for an optimal starting view perspective
        widget.control.spin([1, 0, 0], -np.pi / 4 - 0.25)  # Tilt forward 45 degrees
        widget.control.spin([0, 0, 1], np.pi / 2)  # Rotate around Y-axis 45 degrees
        widget.center(zoom=True)
        widget.control.zoom(0.1)  # Apply zoom factor for closer view

        return widget


def ase2spglib(crystal_ase: Atoms) -> tuple:
    """
    Convert an ASE Atoms object into the tuple format required by spglib.

    This function extracts the lattice, scaled atomic positions, and atomic numbers
    from an ASE Atoms object and returns them in the standard spglib tuple format:
    (lattice, positions, numbers).

    Parameters
    ----------
    crystal_ase : Atoms
        ASE Atoms object representing the crystal structure.

    Returns
    -------
    spglib_crystal : tuple
        A 3-tuple (lattice, positions, numbers) where:
        - lattice : (3, 3) array of lattice vectors
        - positions : (N, 3) array of scaled atomic positions (fractional)
        - numbers : (N,) array of atomic numbers
    """
    lattice = np.asarray(crystal_ase.get_cell()).copy()
    positions = crystal_ase.get_scaled_positions()
    numbers = crystal_ase.get_atomic_numbers()
    spglib_crystal = (lattice, positions, numbers)
    return spglib_crystal


def spglib2ase(spglib_crystal: tuple) -> Atoms:
    """
        Convert a spglib-format crystal tuple into an ASE Atoms object.

        This function takes a tuple of the form (lattice, positions, numbers),
        as used by spglib, and creates a corresponding ASE Atoms object.
    Parameters
        ----------
        spglib_crystal : tuple
            A 3-tuple (lattice, positions, numbers) as returned by spglib,
            where:
            - lattice : (3, 3) array of lattice vectors
            - positions : (N, 3) array of scaled (fractional) positions
            - numbers : (N,) array of atomic numbers

        Returns
        -------
        ase_crystal : Atoms
            ASE Atoms object representing the crystal structure.
    """
    lattice = spglib_crystal[0]
    positions = spglib_crystal[1]
    numbers = spglib_crystal[2]
    ase_crystal = Atoms(
        scaled_positions=positions, numbers=numbers, cell=lattice, pbc=True
    )
    return ase_crystal


def read_spg(file: str) -> tuple:
    """
    Read a crystal structure file and convert it directly to spglib format.

    This function uses ASE's `read()` to load a crystal structure file
    and returns the structure as a tuple in the format required by spglib:
    (lattice, positions, atomic_numbers).

    Parameters
    ----------
    file : str
        Path to the structure file (e.g., CIF, POSCAR, XYZ, etc.) supported by ASE.

    Returns
    -------
    spglib_crystal : tuple
        A 3-tuple (lattice, positions, numbers) where:
        - lattice : (3, 3) array of lattice vectors
        - positions : (N, 3) array of scaled atomic positions (fractional)
        - numbers : (N,) array of atomic numbers
    """
    from ase.io import read

    cryst = read(file)
    spglib_cryst = ase2spglib(cryst)
    return spglib_cryst


def _rot_name(rot: np.ndarray, lattice=np.ndarray):
    """
    Identify the symbolic name and direction of a symmetry operation given by spglib.

    Parameters
    ----------
    rot : ndarray, shape (3, 3)
        Rotation matrix of the symmetry operation.

    lattice : ndarray, shape (3, 3)
        Lattice vectors of the crystal, used to compute Cartesian directions.

    Returns
    -------
    SimpleNamespace
        Object with the following fields:
        - label : str
            Symbolic name of the symmetry element (e.g. 'E', 'C3', 'm', 'I', 'S6')
        - direction_crystal : ndarray or None
            Normalized direction vector in lattice (fractional) coordinates.
        - direction_cartesian : ndarray or None
            Normalized direction vector in Cartesian coordinates.
        - order : int
            Order of the operation (rot^order = identity)
        - det : float
            Determinant of the rotation matrix (+1 or -1)
    """
    E = np.identity(3)
    det = np.linalg.det(rot)
    eigvals, eigvecs = np.linalg.eig(rot)

    # Determine operation order: smallest d such that rot^d ≈ identity
    r = rot.copy()
    d = 1
    while not np.allclose(r, E, atol=1e-6):
        r = r @ rot
        d += 1
        if d > 12:
            break  # fail-safe

    label = "?"
    direction_crystal = None
    direction_cartesian = None

    if np.allclose(rot, E):
        label = "E"
    elif det > 0:  # Proper rotation
        if d > 1:
            idx = np.argmax(np.isclose(eigvals, 1))
            axis = eigvecs[:, idx].real
            direction_crystal = axis / np.linalg.norm(axis)
            direction_cartesian = ut.cryst2cartesian(direction_crystal, lattice)
            direction_cartesian /= np.linalg.norm(direction_cartesian)
            label = f"C{d}"
    elif det < 0:  # Improper rotation
        if d == 2 and np.allclose(rot, -E):
            label = "I"
        else:
            idx = np.argmax(np.isclose(eigvals, -1))
            axis = eigvecs[:, idx].real
            direction_crystal = axis / np.linalg.norm(axis)
            direction_cartesian = ut.cryst2cartesian(direction_crystal, lattice)
            direction_cartesian /= np.linalg.norm(direction_cartesian)
            if d == 1:
                label = "m"
            else:
                label = f"S{d}"

    return SimpleNamespace(
        label=label,
        direction_crystal=direction_crystal,
        direction_cartesian=direction_cartesian,
        order=d,
        det=det,
    )



"""
YAIV | yaiv.convergence
=======================

Tools to collect, organize, and visualize convergence data from ab initio
calculations. Typical use cases include scanning cutoffs, smearing, and k-point grids,
then plotting how total energy, forces, Fermi level, runtime, and memory usage
evolve with these parameters.

Classes
-------
Self_consistent
    Container and utilities for convergence analysis of self‑consistent calculations.
    Provides:
    - read_data(): Recursively scan a folder for outputs and populate convergence data.
    - save_as(): Save the entire Self_consistent object in a .pkl file.
    - from_pkl(): Load a Self_consistent object from a .pkl file.
    - plot(): Plot quantities agains computational parameters for checking convergence.
    - _Analyze:
        Helper for generating multi-panel convergence figures.
        Provides:
        - cutoff(): Generate a multi-panel figure summarizing convergence vs. cutoff.
        - kgrid(): Generate a multi-panel figure summarizing convergence vs. k-grid (and smearing).

Examples
--------
>>> from yaiv.convergence import Self_consistent
>>> analysis = Self_consistent()
>>> analysis.read_data(folder)
>>> analysis.plot("energy", "kgrid", "smearing")
(Figure or the energy evolution respect to the kgrid for different smearings)
>>> analysis.analyze.cutoff()
(Multi-panel figure for analyzis of convergence vs cutoff)
>>> analysis.analyze.kgrid()
(Multi-panel figure for analyzis of convergence vs kgrid)
"""

import glob
from types import SimpleNamespace

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from yaiv.defaults.config import ureg
from yaiv import grep

__all__ = ["Self_consistent"]


class Self_consistent:
    """
    Container and utilities for convergence analysis of self‑consistent calculations.

    This class scans a folder of output files (e.g., QE .pwo files), extracts
    the relevant quantities (cutoff, smearing, k‑grid, runtime, Fermi level,
    memory usage, forces, total energy), stores them in a SimpleNamespace
    with consistent units, and provides simple plotting utilities.

    Attributes
    ----------
    data : SimpleNamespace | None
        Namespace holding arrays for each collected attribute (cutoff, smearing,
        kgrid, runtime, fermi, ram, forces, energy). Each attribute is either a
        NumPy array or a pint.Quantity array with a common unit.

    Methods
    -------
    read_data(folder)
        Recursively scan `folder` for outputs and populate `self.data`.
    plot(x, y, group=None)
        Plot y(x); optionally group by a third attribute `group`.
    save_as("convergence.pkl")
        Save the entire Self_consistent object in a .pkl file.
    from_pkl(filename)
        Load a Self_consistent object from a .pkl file.
    analyze
        Subclass for generating multi-panel convergence figures.
    """

    def __init__(self):
        """
        Initialize an empty Self_consistent object for convergence analysis.

        Use `read_data(folder)` to populate `self.data` from output files.
        """
        self.data = None
        # Full analysis for different DFT parameters
        self.analyze = self._Analyze(self)

    def read_data(self, folder: str):
        """
        Recursively read convergence data from a folder of outputs.

        This scans for files matching `**/*pwo` and extracts from each:
          - cutoff (energy cutoff)
          - smearing
          - kgrid (Monkhorst–Pack grid as (Nk1,Nk2,Nk3))
          - energy (total energy)
          - forces (total atomic force norm)
          - fermi (Fermi energy)
          - ram (peak memory usage if available)
          - runtime (runtime)

        The collected values are converted to consistent units per attribute
        and stored in `self.data` as arrays (or pint.Quantity arrays).

        Parameters
        ----------
        folder : str
            Root directory to search (recursively). For example: "runs/" or "./".

        Notes
        -----
        - This relies on helper parsers in yaiv.grep.
        - Attributes with no data remain None.
        """
        files = glob.glob(folder + "**/*pwo", recursive=True)
        data = SimpleNamespace(
            cutoff=[],
            smearing=[],
            kgrid=[],
            runtime=[],
            fermi=[],
            ram=[],
            forces=[],
            energy=[],
        )
        for file in files:
            data.cutoff.append(grep.cutoff(file))
            data.smearing.append(grep.smearing(file))
            data.kgrid.append(grep.k_grid(file))
            data.runtime.append(grep.runtime(file))
            data.fermi.append(grep.fermi(file))
            data.ram.append(grep.ram(file))
            data.forces.append(grep.atomic_forces(file).total)
            data.energy.append(grep.total_energy(file))
        # Unify units
        for atr in data.__dict__:
            d = data.__getattribute__(atr)
            if len(d) != 0:
                if isinstance(d[0], ureg.Quantity):
                    units = d[0].units
                    new = np.asarray([x.to(units).magnitude for x in d]) * units
                else:
                    new = np.asarray(d)
                data.__setattr__(atr, new)
            else:
                data.__setattr__(atr, None)
        self.data = data

    def save_as(self, filename: str = "convergence.pkl"):
        """
        Save the entire Self_consistent object in a .pkl file.

        Parameters
        ----------
        filename : str, Optional
            Path to the output file. '.pkl' will be added if missing. Default `D_s.pkl`.
        """
        import pickle
        if not filename.endswith(".pkl"):
            filename += ".pkl"
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_pkl(cls, filename: str):
        """
        Load a Self_consistent object from a .pkl file.

        Parameters
        ----------
        filename : str
            Path to file created with `save_as`.

        Returns
        -------
        Self_consistent
            Fully reconstructed Self_consistent object.
        """
        import pickle
        with open(filename, "rb") as f:
            obj = pickle.load(f)
        return obj

    def plot(
        self,
        x: str,
        y: str,
        group: str = None,
        ax: Axes = None,
        **kwargs,
    ) -> (Axes, Axes):
        """
        Plot y vs x with optional grouping.

        Parameters
        ----------
        x : str
            Name of the attribute in `self.data` for the x‑axis
            If x is "kgrid", the product Nk1*Nk2*Nk3 is plotted.
        y : str
            Name of the attribute in `self.data` for the y‑axis.
        group : str, optional
            Name of an attribute in `self.data` used to group the data into curves
            (e.g., group="smearing"). Each unique group value produces a separate line.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on. If None, a new figure and axes are created.
        **kwargs : dict
            Additional keyword arguments passed to `ax.plot()`.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the plotted data.

        Raises
        ------
        NameError
            If `x`, `y`, or `group` (when provided) do not exist in `self.data`.
        """
        # Get X, Y
        attributes = self.data.__dict__.keys()
        if x not in attributes:
            raise NameError(f"{x} attribute not present at self.data.\n {attributes}")
        else:
            X = self.data.__getattribute__(x)
        if y not in attributes:
            raise NameError(f"{y} attribute not present at self.data.\n {attributes}")
        else:
            Y = self.data.__getattribute__(y)

        # Special case: kgrid on x-axis -> use total number of k-points (product)
        if isinstance(X[0], np.ndarray):
            X = np.asarray([np.prod(grid) for grid in X])

        # Prepare groups
        if group is not None:
            if group not in attributes:
                raise NameError(
                    f"{group} attribute not present at self.data.\n {attributes}"
                )
            else:
                Z = self.data.__getattribute__(group)
            if isinstance(Z, ureg.Quantity):
                groups = np.unique(Z.magnitude, axis=0) * Z.units
            else:
                groups = np.unique(Z, axis=0)

        # Create fig if necessary
        if ax is None:
            fig, ax = plt.subplots()

        # Plot grouped or ungrouped
        if group is not None:
            for g in groups:
                indices = []
                for i, z in enumerate(Z):
                    if np.all(z == g):
                        indices.append(i)
                Ysort = Y[indices][np.argsort(X[indices])]
                Xsort = X[indices][np.argsort(X[indices])]
                ax.plot(Xsort, Ysort, ".-", label=str(g), **kwargs)
        else:
            Ysort = Y[np.argsort(X)]
            Xsort = X[np.argsort(X)]
            ax.plot(X, Y, ".-", **kwargs)

        # Decorations
        if isinstance(X, ureg.Quantity):
            ax.set_xlabel(f"{x} ({X.units})")
        else:
            ax.set_xlabel(f"{x}")
        if isinstance(Y, ureg.Quantity):
            ax.set_ylabel(f"{y} ({Y.units})")
        else:
            ax.set_xlabel(f"{y}")
        if group is not None:
            ax.legend()

        plt.tight_layout()
        return ax

    class _Analyze:
        """
        Analysis helper for generating multi-panel convergence figures.

        This sub-class produces full-page summary plots to visualize how key
        quantities (total energy, forces, Fermi level, runtime, memory) change
        with respect to specific convergence parameters (e.g., cutoff, k-grid,
        smearing). It leverages the parent `Self_consistent` container to access
        collected data and the `plot` method for consistent plotting.

        Attributes
        ----------
        parent : Self_consistent
            Reference to the parent convergence container from which data are obtained.

        Methods
        -------
        cutoff()
            Generate a multi-panel figure summarizing convergence vs. cutoff.
        kgrid()
            Generate a multi-panel figure summarizing convergence vs. k-grid (and smearing).
        """

        def __init__(self, parent):
            """
            Initialize the analysis helper.

            Parameters
            ----------
            parent : Self_consistent
                The parent container holding convergence data (from `read_data`) and
                the plotting utility (`plot`). Must be already populated for meaningful
                output.
            """
            self._p = parent  # reference to the parent convergence class

        def cutoff(self):
            """
            Generate a multi-panel figure summarizing convergence vs. cutoff.

            The figure shows, for each quantity (energy, forces, Fermi level, runtime,
            memory):
              - Left column: y vs cutoff, grouped by k-grid (total Nk = Nk1*Nk2*Nk3).
              - Right column: y vs k-grid (total Nk), grouped by cutoff.

            Notes
            -----
            - Requires that `self._p.data` is populated via `read_data`.
            """
            fig, axes = plt.subplots(5, 2, figsize=(10.5, 14.5))
            ax = axes.flatten()
            x = "cutoff"
            z = "kgrid"
            for i, y in enumerate(["energy", "forces", "fermi", "runtime", "ram"]):
                self._p.plot(x, y, z, ax[2 * i])
                self._p.plot(z, y, x, ax[2 * i + 1])
            return fig, axes

        def kgrid(self):
            """
            Generate a multi-panel figure summarizing convergence vs. k-grid (and smearing).

            The figure shows, for each quantity (energy, forces, Fermi level, runtime,
            memory):
              - Left column: y vs smearing, grouped by k-grid (total Nk = Nk1*Nk2*Nk3).
              - Right column: y vs k-grid (total Nk), grouped by smearing.

            Notes
            -----
            - Requires that `self._p.data` is populated via `read_data`.
            """
            fig, axes = plt.subplots(5, 2, figsize=(10.5, 14.5))
            ax = axes.flatten()
            x = "smearing"
            z = "kgrid"
            for i, y in enumerate(["energy", "forces", "fermi", "runtime", "ram"]):
                self._p.plot(x, y, z, ax[2 * i])
                self._p.plot(z, y, x, ax[2 * i + 1])
            return fig, axes



"""
YAIV | yaiv.grep
================

This module provides text-scraping utilities for extracting (grepping) structural and spectral
information from first-principles calculation outputs. It supports common DFT packages
such as Quantum ESPRESSO and VASP.

The functions in this module perform low-level parsing (i.e., grepping) of data such as:

- Electronic eigenvalues and k-points
- Phonon frequencies and paths
- Lattice vectors and stress tensors
- Number of electrons and total energies
- Fermi level and reciprocal space paths
- Computational cost (runtime, RAM)
- DFT parameters as cutoff, smearing or K-grid

Supported formats include:
- Quantum ESPRESSO output/input: `pw.x`, `ph.x`, `bands.in`, `projwfc.x`, `matdyn.x`, `.xml`
- VASP output: `OUTCAR`, `EIGENVAL`, `KPOINTS`, `PROCAR`

Functions
---------
electron_num(file)
    Greps the number of electrons from a QE or VASP output file.

lattice(file, alat=False)
    Extracts the lattice vectors from outputs. Optionally in internal units (alat).

fermi(file)
    Extracts the Fermi energy from an output file.

total_energy(file, decomposition=False)
    Extracts the total free energy, optionally returning its internal decomposition (QE).

stress_tensor(file)
    Extracts the total stress tensor from a QE or VASP output file.

kpath(file, labels=True)
    Extracts the k-point path in reciprocal space, optionally with high-symmetry labels.

kpointsEnergies(file)
    Greps the k-points, energies, kpoint-weights  and orbital projections (if available) for different file kinds.

kpointsFrequencies(file)
    Extracts k-points and phonon frequencies from Quantum ESPRESSO phonon outputs.

dyn_file(file)
    Reads a QE `.dyn` file and returns vibrational data (q-point (2π/Å), lattice (Å), frequencies, ...)

dyn_q(q_cryst, results_ph_path, qe_format=True)
    Locates and reads `.dyn*` file for a given q-point, returning the full dynamical matrix (3Nx3N).

symmetries(file)
    Grep symmetry operations and return them as rotation/translation pairs.

cutoff(file)
    Greps the cutoff energy from a variety of filetypes.

smearing(file)
    Greps the smearing used in calculation from a variety of filetypes.

runtime(file)
    Greps the computational runtime from a variety of filetypes.

ram(file)
    Greps the RAM needed in the computation from a variety of filetypes.

k_grid(file)
    Greps the k-grid from a variety of filetypes.

atomic_forces(file)
    Greps the atomic forces from a variety of filetypes.

Private Utilities
-----------------
_filetype(file)
    Heuristically determines the type of simulation output file (QE, VASP, etc.).

_OrbitalProjectionContainer:
    Container for orbital-resolved projection matrices.

_Symmetry:
    Container for symmetry operations.

_Qe_xml:
    Minimal reader for Quantum ESPRESSO XML output files.

_find_dyn_file(q_cryst, results_ph_path)
    Internal helper that searches `.dyn*` files matching the specified q-point.

Examples
--------
>>> from yaiv grep import grep
>>> spectrum = grep.kpointsEnergies("OUTCAR")
>>> spectrum.eigenvalues.shape
(100, 32)

>>> from yaiv grep import grep
>>> a = grep.lattice("qe.out")
>>> a.shape
(3, 3)

See Also
--------
yaiv.spectrum : Data container and plotter for eigenvalue spectra
yaiv.utils    : Basis universal utilities
"""

import re
import warnings
from types import SimpleNamespace

import numpy as np

from yaiv.defaults.config import ureg
from yaiv import utils as ut

__all__ = [
    "electron_num",
    "lattice",
    "fermi",
    "total_energy",
    "stress_tensor",
    "kpath",
    "kpointsEnergies",
    "kpointsFrequencies",
    "dyn_file",
    "dyn_q",
    "symmetries",
    "cutoff",
    "smearing",
    "k_grid",
    "atomic_forces",
    "runtime",
    "ram",
]


def _filetype(file: str) -> str:
    """
    Detects the filetype of the provided file.

    Parameters
    ----------
    file : str
        Filepath for the file to analyze.

    Returns
    -------
    filetype : str
        Detected filetype (None if not filetype is detected).
    """
    filetype = None
    with open(file, "r") as lines:
        for line in lines:
            line = line.strip().lower()
            if re.search(r"calculation.*scf|calculation.*nscf", line):
                filetype = "qe_scf_in"
                break
            elif "program pwscf" in line:
                filetype = "qe_scf_out"
                break
            elif "program phonon" in line:
                filetype = "qe_ph_out"
                break
            elif "calculation" in line and "bands" in line:
                filetype = "qe_bands_in"
                break
            elif "flfrc" in line:
                filetype = "matdyn_in"
                break
            elif "dynamical matrix" in line:
                filetype = "qe_dyn"
                break
            elif "&plot nbnd=" in line:
                filetype = "qe_freq_out"
                break
            elif "projwave" in line:
                filetype = "qe_proj_out"
                break
            elif "procar" in line:
                filetype = "procar"
                break
            elif "vasp" in line:
                filetype = "outcar"
                break
            elif len(line.split()) == 4 and all(x.isdigit() for x in line.split()):
                filetype = "eigenval"
                break
            elif "line-mode" in line:
                filetype = "kpath"
                break
            elif ("direct" in line and "directory" not in line) or "cartesian" in line:
                filetype = "poscar"
                break
            elif "espresso xml" in line or "espresso xsi" in line:
                filetype = "qe_xml"
                break
    return filetype


class _OrbitalProjectionContainer:
    """
    Container for orbital-resolved projection matrices.

    Supports projections stored from:
      - POSCAR-style (VASP-like): keys (ion, l, ml, M)
      - QE (Quantum ESPRESSO) SOC projwfc: keys (ion, l, j, mj, wfc)
      - QE (Quantum ESPRESSO) scalar projwfc: keys (ion, l, m, wfc)

    Each entry maps a quantum-number tuple → projection matrix of shape (nkpts, nbnds).

    Attributes
    ----------
    _data : dict
        Internal mapping from quantum-number tuples to projection matrices.
        - POSCAR format: (ion, l, m, M) → np.ndarray (nkpts, nbnds)
        - QE SOC format:     (ion, l, j, mj, wfc) → np.ndarray (nkpts, nbnds)
        - QE scalar format:  (ion, l, m, wfc) → np.ndarray (nkpts, nbnds)
    _code : str
        Code according to which the code is formated. Options are "quantum_espresso" and "vasp".

    Methods
    -------
    add_poscar(ion, l, m, M, matrix)
        Store a projection matrix in POSCAR format.
    add_qe(ion, l, wfc, matrix, m, j, mj)
        Store a projection matrix in QE projwfc format.
    __call__(...)
        Sum over matching entries with flexible indexing across both formats.
    __repr__()
        Compact summary of stored channels (POSCAR or QE).
    """

    def __init__(self, code=str):
        """
        Initialize an empty orbital projection container.

        Parameters
        ----------
        code : str
            Code according to which the code is formated. Options are "quantum_espresso" and "vasp".
        """
        self._data = {}
        self._code = code

    def add_poscar(self, ion: int, l: int, m: int, M: int, matrix: np.ndarray):
        """
        Store a projection matrix for a specific channel in POSCAR format.

        Parameters
        ----------
        ion : int
            Ion index the projection refers to.
        l : int
            Orbital angular momentum (0=s, 1=p, 2=d, ...).
        m : int
            Magnetic quantum number (−l..+l).
        M : int
            Magnetization channel: 0 for total (|M|), 1,2,3 for x,y,z components.
        matrix : np.ndarray, shape (nkpts, nbnds)
            Projection matrix.
        """
        self._data[(ion, l, m, M)] = matrix

    def add_qe(
        self,
        ion: int,
        l: int,
        wfc: int,
        matrix: np.ndarray,
        m: int = None,
        j: float = None,
        mj: float = None,
    ):
        """
        Store a projection matrix for a specific QE projwfc channel.

        Parameters
        ----------
        ion : int
            Ion index the projection refers to.
        l : int
            Orbital angular momentum (0=s, 1=p, 2=d, ...).
        wfc : int
            Wavefunction index (QE projwfc channel identifier).
        matrix : np.ndarray, shape (nkpts, nbnds)
            Projection matrix.
        m : int, optional
            QE real-harmonic component index (1..2*l+1). Not a magnetic quantum number.
            Used in scalar mode.
        j : float, optional
            Total angular momentum (e.g., 0.5, 1.5, ...). Used in SOC mode.
        mj : float, optional
            Projection of total angular momentum (−j..+j, step 1). Used in SOC mode.

        Raises
        ------
        ValueError
            If an invalid combination of (m, j, mj) is provided or values are out of range.
        """
        scalar_mode = (m is not None) and (j is None) and (mj is None)
        soc_mode = (m is None) and (j is not None) and (mj is not None)

        if scalar_mode:
            self._data[(ion, l, m, wfc)] = matrix
        elif soc_mode:
            self._data[(ion, l, j, mj, wfc)] = matrix
        else:
            raise ValueError(
                "Invalid QE projection description. "
                "Provide either (m) for scalar calculations "
                "or (j, mj) for spin–orbit calculations."
            )

    def __repr__(self) -> str:
        keys = np.array(tuple(self._data.keys()), dtype=object)
        if self._code == "quantum_espresso":
            ions = int(max(int(k[0]) for k in keys) + 1)
            l_by_ion = {
                ion: sorted({int(k[1]) for k in keys if int(k[0]) == ion})
                for ion in range(ions)
            }
            # QE_SOC: keys of length 5 -> (ion, l, j, mj, wfc)
            if len(keys[0]) == 5:
                return (
                    f"_OrbitalProjectionContainer("
                    f"ions={ions}, \n"
                    f"l={l_by_ion} \n"
                    f"shape=(ion, l, j, m_j, wfc))"
                )
            # QE_scalar: keys of length 5 -> (ion, l, m, wfc)
            elif len(keys[0]) == 4:
                return (
                    f"_OrbitalProjectionContainer("
                    f"ions={ions}, \n"
                    f"l={l_by_ion} \n"
                    f"shape=(ion, l, m[1,1+2l], wfc)"
                )
        # POSCAR: keys of length 4 -> (ion, l, m, M)
        elif self._code == "vasp":
            keys = np.array(keys, dtype=int)
            ions = np.max(keys[:, 0]) + 1
            l = sorted(set(int(x) for x in keys[:, 1]))
            Mmax = int(np.max(keys[:, 3]))
            Mstr = f"[0-{Mmax}]" if Mmax > 0 else "0"
            return (
                f"_OrbitalProjectionContainer(ions={ions}, l={l}, m_l=[-l,l], M={Mstr}\n"
                f"shape=(ion, l, m_l, M))"
            )
        else:
            return f"_OrbitalProjectionContainer(keys_shape={len(keys[0])}, entries={len(keys)})"

    def __call__(
        self,
        ion: int | range | tuple | list | slice = slice(None),
        l: int | range | tuple | list | slice = slice(None),
        m: int | range | tuple | list | slice = slice(None),
        M: int | range | tuple | list | slice = 0,
        j: float | tuple | list | slice = slice(None),
        mj: float | tuple | list | slice = slice(None),
        wfc: int | range | tuple | list | slice = slice(None),
    ) -> np.ndarray:
        """
        Sum all projection matrices matching the query.

        This method supports both POSCAR-style and QE-style keys transparently.
        Provide the subset of quantum numbers you want to match; any argument can be
        a scalar, list/tuple/range, or slice(None) to select all.

        Parameters
        ----------
        ion : int | list | range | slice, optional
            Ion index or indices. Default selects all ions.
        l : int | list | range | slice, optional
            Orbital angular momentum(s). Default selects all.
        m : int | list | range | slice, optional
            Magnetic quantum number(s) (−l..+l), for POSCAR data.
            Real-harmonic component index [1,1+2l], for scalar QE.
        M : int | list | range | slice, optional (POSCAR)
            Magnetization channel (0 total, 1,2,3 for x,y,z). Default 0.
        j : float | list | tuple | slice, optional (QE)
            Total angular momentum(s). Only used for QE data.
        mj : float | list | tuple | slice, optional (QE)
            m_j values (projection of J). Only used for QE data.
        wfc : int | list | range | slice, optional (QE)
            Wavefunction channel indices. Only used for QE data.

        Returns
        -------
        np.ndarray
            Sum of all projection matrices matching the query.

        Raises
        ------
        KeyError
            If no entries match the given query.

        Examples
        --------
        # POSCAR-style:
        proj(ion=0, l=1, m=0, M=1)               # single matrix
        proj(ion=[0,1,2], l=1, m=0, M=1)         # sum over several ions
        proj(ion=slice(None), l=1, m=0, M=1)     # sum over all ions

        # QE-style:
        proj(ion=0, l=1, j=1.5, mj=0.5, wfc=3)   # single matrix
        proj(ion=0, l=1, j=slice(None))          # sum over all J
        proj(ion=[0,1], l=[0,1], j=[0.5,1.5])    # sum over subsets
        """
        keys = np.array(tuple(self._data.keys()), dtype=object)
        # QE keys length 5
        if self._code=="quantum_espresso":
            if len(keys[0]) == 5:
                query = (ion, l, j, mj, wfc)
            elif len(keys[0]) == 4:
                query = (ion, l, m, wfc)
        elif self._code=="vasp":
            # POSCAR keys length 4
            query = (ion, l, m, M)
        else:
            raise KeyError("Unknown code in container")

        def match(val, key):
            if isinstance(val, slice):
                return True
            elif isinstance(val, (list, tuple, set, range)):
                return key in val
            else:
                return key == val

        # Direct (single) match returns the stored matrix
        if all(not isinstance(q, (list, tuple, set, range, slice)) for q in query):
            return self._data[query]

        # Sum all matching matrices
        selected = [
            mat
            for key, mat in self._data.items()
            if all(match(q, k) for q, k in zip(query, key))
        ]
        if not selected:
            raise KeyError(f"No projection matches query: {query}")
        return sum(selected)


class _Symmetry:
    def __init__(
        self,
        R: np.ndarray,
        t: np.ndarray = None,
        units: ureg.Quantity = ureg.dimensionless,
    ):
        """
        Container for symmetry operations with rotation and (optional) translation in a given basis.

        Parameters
        ----------
        R : np.ndarray
            Rotation matrix in the current coordinate system (matches `units`).
        t : np.ndarray, optional
            Translation (shift) vector in the current coordinate system. Default is zero.
        units : ureg.Quantity, optional.
            Units of `t` (and the basis in which `R` is expressed). Typically
            crystal (e.g., ureg('crystal')) or Cartesian length (e.g., ureg.meter).

        Methods
        -------
        to_cartesian(lattice)
            Convert this symmetry from crystal to Cartesian coordinates.
        to_crystal(lattice)
            Convert this symmetry from Cartesian to crystal coordinates.

        Notes
        -----
        - This container does not assume or enforce a specific convention
          (row/column) beyond storing R and t consistently in the same basis.
        - Use `to_cartesian` / `to_crystal` to convert both R and t between bases.
        """

        self.R = R
        if t is None:
            t = np.zeros(R.shape[0])
        self.t = t
        self.units = units

    def __repr__(self) -> str:
        return f"_Symmetry \n {self.R}, {self.t}, {self.units}"

    # ------------- change of basis ---------
    def to_cartesian(self, lattice: np.ndarray | ureg.Quantity) -> "_Symmetry":
        """
        Convert _Symmetry object to Cartesian coordinates.

        Parameters
        ----------
        lattice : np.ndarray | ureg.Quantity
            Direct lattice vectors (rows) in Cartesian units (e.g., Å). If
            given as a Quantity, units are used; otherwise assumed Cartesian length

        Returns
        -------
        Sym_new : _Symmetry
            New _Symmetry object in Cartesian coord.

        Raises
        ------
        ValueError
            If the current `units` are not crystal units.
        """
        if self.units.dimensionality != ureg.crystal.dimensionality:
            raise ValueError(
                f"Only can chage to Cartesian coord from crystal units. Not {self.units} units."
            )
        t = ut.cryst2cartesian(self.t * self.units, lattice)
        if isinstance(lattice, ureg.Quantity):
            lattice = lattice.magnitude
        cryst2cartesian = lattice.T
        cartesian2cryst = ut.invQ(cryst2cartesian)
        R = cryst2cartesian @ self.R @ cartesian2cryst
        return _Symmetry(R, t.magnitude, t.units)

    def to_crystal(self, lattice: np.ndarray | ureg.Quantity) -> "_Symmetry":
        """
        Convert _Symmetry object to crystal coordinates.

        Parameters
        ----------
        lattice : np.ndarray | ureg.Quantity
            Direct lattice vectors (rows) in Cartesian units (e.g., Å). If
            given as a Quantity, units are used; otherwise assumed Cartesian length

        Returns
        -------
        Sym_new : _Symmetry
            New _Symmetry object in crystal coord.

        Raises
        ------
        ValueError
            If the current `units` are not in lenght dimnesion.
        """
        if self.units.dimensionality != ureg.meter.dimensionality:
            raise ValueError(
                f"Only can chage to crystal coord from length units. Not {self.units} units."
            )
        t = ut.cartesian2cryst(self.t * self.units, lattice)
        if isinstance(lattice, ureg.Quantity):
            lattice = lattice.magnitude
        cryst2cartesian = lattice.T
        cartesian2cryst = ut.invQ(cryst2cartesian)
        R = cartesian2cryst @ self.R @ cryst2cartesian
        return _Symmetry(R, t.magnitude, t.units)


class _Qe_xml:
    """
    Minimal reader for Quantum ESPRESSO XML output files.

    Provides utilities to extract common physical quantities.

    Notes
    -----
    - Units: numerical values in the XML are in Hartree atomic units unless
      otherwise specified. Returned values are wrapped in `ureg.Quantity`.
    """

    def __init__(self, file):
        """
        Initialize a QE XML reader.

        Parameters
        ----------
        file : str or Path
            Path to the Quantum ESPRESSO XML file.

        Raises
        ------
        NotImplementedError
            If the file type is not recognized as a QE XML file.
        """
        import xml.etree.ElementTree as ET

        if _filetype(file) == "qe_xml":
            tree = ET.parse(file)
            self.root = tree.getroot()
        else:
            raise NotImplementedError("Unsupported filetype")

    def electron_num(self) -> int:
        """
        Greps the number of electrons.

        Returns
        -------
        num_elec : int
            Number of electrons.
        """
        elec = self.root.find(".//nelec")
        return int(float(elec.text))

    def lattice(self) -> np.ndarray:
        """
        Greps the lattice vectors.

        Returns
        -------
        lattice : np.ndarray
            3x3 array of lattice vectors with attached units (ureg.Quantity).
        """
        cell = self.root.find(".//cell")
        lattice = []
        for line in cell:
            v = [float(x) for x in line.text.split()]
            lattice += [v]
        lattice = np.array(lattice) * ureg.bohr
        return lattice

    def fermi(self) -> float:
        """
        Greps the Fermi energy.

        Returns
        -------
        E_f : float
            Fermi energy with attached units (ureg.Quantity).
        """
        fermi = self.root.find(".//fermi_energy")
        return float(fermi.text) * ureg.hartree

    def total_energy(self, decomposition: bool = False) -> float | SimpleNamespace:
        """
        Greps the total free energy or it's decomposition.

        Parameters
        ----------
        decomposition : bool, optional
            If True an energy decomposition is returned instead. Default is False.

        Returns
        -------
        energy : float | SimpleNamespace
            If decomposition is False a single float with the free energy is returned.
            If decomposition is True a namespace with the following attributes is returned:
                -  F            -> Total Free energy
                - -TS           -> Smearing contribution
                -  U (= F+TS)   -> Internal energy
                    -  U_one_electron
                    -  U_hartree
                    -  U_exchange-correlational
                    -  U_ewald
        """
        lines = self.root.find(".//total_energy")
        etot = float(lines.find(".//etot").text) * ureg.hartree
        eband = float(lines.find(".//eband").text) * ureg.hartree
        ehart = float(lines.find(".//ehart").text) * ureg.hartree
        vtxc = float(lines.find(".//vtxc").text) * ureg.hartree
        etxc = float(lines.find(".//etxc").text) * ureg.hartree
        ewald = float(lines.find(".//ewald").text) * ureg.hartree
        demet = float(lines.find(".//demet").text) * ureg.hartree
        energy = SimpleNamespace(
            F=etot,
            TS=demet,
            U=etot - demet,
            U_one_electron=etot - demet - ehart - etxc - ewald,
            U_hartree=ehart,
            U_xc=etxc,
            U_ewald=ewald,
        )
        if decomposition:
            return energy
        else:
            return energy.F

    def kpointsEnergies(self) -> SimpleNamespace:
        """
        Grep the kpoints, energies and kpoint-weights.

        Returns
        -------
        SimpleNamespace : SimpleNamespace
            SimpleNamespace class with the following attributes:
            - energies : np.ndarray
                List of energies, each row corresponds to a particular k-point.
            - kpoints : np.ndarray
                List of k-points.
            - weights : np.ndarray
                List of kpoint-weights.
        """
        KPOINTS, WEIGHTS, ENERGIES = [], [], []
        ks_energies = self.root.findall(".//ks_energies")
        for elem in ks_energies:
            # Get kpoint and weights
            kpoint = elem.find(".//k_point")
            w = float(kpoint.attrib["weight"])
            k = [float(x) for x in kpoint.text.split()]
            KPOINTS += [k]
            WEIGHTS += [w]
            # Get energies
            E = [float(x) for x in elem.find(".//eigenvalues").text.split()]
            ENERGIES += [E]
        return SimpleNamespace(
            energies=ENERGIES * ureg.hartree,
            kpoints=KPOINTS * (ureg._2pi / ureg.alat),
            weights=np.array(WEIGHTS),
        )

    def symmetries(self) -> list[SimpleNamespace]:
        """Grep symmetry operations from the QE XML and return them as rotation/translation pairs.

        This reads all <symmetry> elements, extracts the 3×3 rotation matrix from
        <rotation> and the fractional translation from <fractional_translation>,
        and returns a list of objects with fields:
          - R: np.ndarray shape (3, 3)
          - t: lenght-3 translation vector
          - units: ureg.Quantity with the units in which {R|t} is given.

        Returns
        -------
        symmetries : list[SimpleNamespace]
            Each entry contains R (3×3 rotation matrix) and t (translation
            vector) and units.
        """
        OUT = []
        symmetries = self.root.findall(".//symmetry")
        for elem in symmetries:
            rotation = elem.find(".//rotation")
            R = np.fromstring(rotation.text, sep=" ").reshape(3, 3)
            translation = elem.find(".//fractional_translation")
            if translation is not None:
                t = np.array([float(x) for x in translation.text.split()])
            else:
                t = np.zeros(3)
            OUT.append(_Symmetry(R=R, t=t, units=ureg.crystal))
        return OUT

    def cutoff(self) -> ureg.Quantity:
        """
        Greps the cutoff energy.

        Returns
        -------
        cutoff : ureg.Quantity
            Cutoff energy with attached units (ureg.Quantity).
        """
        cutoff = float(self.root.find(".//ecutwfc").text) * ureg.Ry
        return cutoff

    def smearing(self) -> ureg.Quantity:
        """
        Greps the smearing.

        Returns
        -------
        smearing : ureg.Quantity
            Smearing with attached units (ureg.Quantity).
        """
        smearing = float(self.root.find(".//smearing").attrib["degauss"]) * ureg.Ry
        return smearing

    def runtime(self) -> ureg.Quantity:
        """
        Greps the computational runtime.

        Returns
        -------
        runtime : ureg.Quantity
            Computational runtime with attached units (ureg.Quantity).
        """
        lines = self.root.find(".//timing_info")
        runtime = float(lines.find(".//total").find(".//cpu").text)
        return runtime * ureg.second

    def k_grid(self) -> ureg.Quantity:
        """
        Greps the k-grid from a variety of filetypes.

        Returns
        -------
        k_grid : list(int)
            K-grid used in the computation.
        """
        lines = self.root.find(".//k_points_IBZ").find(".//monkhorst_pack")
        kgrid = [
            int(lines.attrib["nk1"]),
            int(lines.attrib["nk2"]),
            int(lines.attrib["nk3"]),
        ]
        return kgrid


def electron_num(file: str) -> int:
    """
    Greps the number of electrons.

    It supports different filetypes as Quantum Espresso or VASP outputs.

    Parameters
    ----------
    file : str
        File from which to extract the electron number, it currently supports:
        - QuantumEspresso `xml` or pw.x output.
        - VASP OUTCAR.

    Returns
    -------
    num_elec : int
        Number of electrons.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The number of electrons was not found in the provided file.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_scf_out":
            for line in lines:
                if "number of electrons" in line:
                    num_elec = int(float(line.split()[4]))
                    break
        elif filetype == "qe_xml":
            num_elec = _Qe_xml(file).electron_num()
        elif filetype == "outcar":
            for line in lines:
                if "NELECT" in line:
                    num_elec = int(float(line.split()[2]))
                    break
        elif filetype == "eigenval":
            for line in lines:
                if len(line.split()) == 3:
                    num_elec = int(line.split()[0])
                    break
        else:
            raise NotImplementedError("Unsupported filetype")
        if "num_elec" not in locals():
            raise NameError("Number of electrons not found.")
    return num_elec


def lattice(file: str, alat: bool = False) -> ureg.Quantity:
    """
    Greps the lattice vectors from various outputs.

    Parameters
    ----------
    file : str
        File from which to extract the lattice.
    alat : bool, optional
        Whether to return lattice in internal units (alat). Default is False.

    Returns
    -------
    lattice : np.ndarray
        3x3 array of lattice vectors with attached units (ureg.Quantity).
        Units will be 'alat' if the `alat` flag is True.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    """
    from ase import io

    filetype = _filetype(file)
    READ = False

    if filetype == "qe_xml":
        lattice = _Qe_xml(file).lattice()
        if alat:
            return lattice / np.linalg.norm(lattice[0]) * ureg.alat
        else:
            return lattice
    elif filetype == "qe_ph_out":
        with open(file, "r") as lines:
            for line in lines:
                if "lattice parameter" in line:
                    line = line.split()
                    ALAT = float(line[4]) * ureg.bohr  # lattice parameter in Bohr
                elif re.search("crystal axes", line, flags=re.IGNORECASE):
                    READ = True
                    lattice = []
                elif READ:
                    values = line.split()
                    vec = np.array(
                        [float(values[3]), float(values[4]), float(values[5])]
                    )
                    lattice.append(vec)
                    if len(lattice) == 3:
                        break
        if alat:
            return lattice * ureg.alat  # lattice in lattice units
        else:
            # Convert alat to Å
            lattice = np.array(lattice) * ALAT.to("angstrom")
            return lattice

    elif filetype == "qe_dyn":
        with open(file, "r") as lines:
            for line in lines:
                if not READ and len(line.split()) == 9:
                    ALAT = float(line.split()[3]) * ureg("bohr/alat")
                elif "Basis vectors" in line:
                    READ = True
                    lattice = []
                elif READ:
                    vec = [float(x) for x in line.split()]
                    lattice.append(vec)
                    if len(lattice) == 3:
                        lattice = np.array(lattice) * ureg.alat
                        break
        if alat:
            return lattice
        else:
            return (lattice * ALAT).to("angstrom")

    elif filetype == "qe_proj_out":
        raise NotImplementedError(
            "Unsupported filetype: ASE is not handling it correctly"
        )
    else:
        # Fallback to ASE
        try:
            lattice = io.read(file).cell  # (3, 3) in Å
        except io.formats.UnknownFileTypeError:
            raise NotImplementedError(
                "Unsupported filetype: ASE is not handling it correctly"
            )
        if alat:
            # Normalize to lattice units
            a_norm = np.linalg.norm(lattice[0])
            return (lattice / a_norm) * ureg("alat")
        else:
            return lattice * ureg.angstrom


def fermi(file: str) -> ureg.Quantity:
    """
    Greps the Fermi energy from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the Fermi energy.

    Returns
    -------
    E_f : ureg.Quantity
        Fermi energy with attached units (ureg.Quantity).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The Fermi energy was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            E_f = _Qe_xml(file).fermi()
        elif filetype == "qe_scf_out":
            for line in reversed(list(lines)):
                # If smearing is used
                if "Fermi energy is" in line:
                    E_f = float(line.split()[4])
                    break
                # If fixed occupations is used
                if "highest occupied" in line:
                    if "unoccupied" in line:
                        split = line.split()
                        E1, E2 = float(split()[6]), float(split()[7])
                        # Fermi level between the unoccupied and occupied bands
                        E_f = E1 + (E2 - E1) / 2
                    else:
                        E_f = float(line.split()[4])
                    break
            E_f *= ureg("eV")
        elif filetype == "outcar":
            for line in reversed(list(lines)):
                if "E-fermi" in line:
                    E_f = float(line.split()[2]) * ureg("eV")
                    break
        else:
            raise NotImplementedError("Unsupported filetype")
    if "E_f" not in locals():
        raise NameError("Fermi energy not found.")
    return E_f


def total_energy(file: str, decomposition: bool = False) -> float | SimpleNamespace:
    """
    Greps the total free energy or it's decomposition.

    Parameters
    ----------
    file : str
        File from which to extract the energy.
    decomposition : bool, optional
        If True an energy decomposition is returned instead. Default is False.

    Returns
    -------
    energy : float | SimpleNamespace
        If decomposition is False a single float with the free energy is returned.
        If decomposition is True a namespace with the following attributes is returned:
            -  F            -> Total Free energy
            - -TS           -> Smearing contribution
            -  U (= F+TS)   -> Internal energy
                -  U_one_electron
                -  U_hartree
                -  U_exchange-correlational
                -  U_ewald
                -  U_paw -> One center paw contribution
                -  U_correction -> Empirical correction (VdW for example.)

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The energy was not found in the provided file.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            energy = _Qe_xml(file).total_energy(decomposition)
        elif filetype == "qe_scf_out":
            data = {}
            for line in reversed(list(lines)):
                if "!" in line:
                    data["F"] = float(line.split()[4]) * ureg("Ry")
                    break
                elif "smearing contrib" in line:
                    data["TS"] = float(line.split()[4]) * ureg("Ry")
                elif "internal energy" in line:
                    data["U"] = float(line.split()[4]) * ureg("Ry")
                elif "one-electron" in line:
                    data["U_one_electron"] = float(line.split()[3]) * ureg("Ry")
                elif "hartree contribution" in line:
                    data["U_hartree"] = float(line.split()[3]) * ureg("Ry")
                elif "xc contribution" in line:
                    data["U_xc"] = float(line.split()[3]) * ureg("Ry")
                elif "ewald" in line:
                    data["U_ewald"] = float(line.split()[3]) * ureg("Ry")
                elif "Dispersion Correction" in line:
                    data["U_correction"] = float(line.split()[3]) * ureg("Ry")
                elif "one-center paw" in line:
                    data["U_paw"] = float(line.split()[4]) * ureg("Ry")
                elif "convergence NOT achieved" in line:
                    raise NameError(f"Convergence not achieved in {file}")
            if decomposition:
                energy = SimpleNamespace(**data)
            else:
                energy = data["F"]
        elif filetype == "outcar":
            for line in reversed(list(lines)):
                if "sigma->" in line:
                    l = line.split()
                    energy = float(l[-1])
                    break
            energy = energy * ureg("eV").to("Ry")
        else:
            raise NotImplementedError("Unsupported filetype")
    if "energy" not in locals():
        raise NameError("Total energy not found.")
    return energy


def stress_tensor(file: str) -> ureg.Quantity:
    """
    Greps the total stress tensor.

    Parameters
    ----------
    file : str
        File from which to extract the stress tensor.

    Returns
    -------
    stress : ureg.Quantity
        Stress tensor with attached units.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The energy was not found in the provided file.
    """
    filetype = _filetype(file)
    READ = False
    stress = []
    with open(file, "r") as lines:
        if filetype == "qe_scf_out":
            for line in lines:
                if READ == True:
                    vec = np.array([float(x) for x in line.split()[:3]])
                    stress.append(vec)
                    if np.shape(stress) == (3, 3):
                        break
                elif re.search("total.*stress", line):
                    READ = True
            stress = np.array(stress) * (ureg("Ry") / ureg("bohr") ** 3).to("kbar")
        elif filetype == "outcar":
            for line in lines:
                if "in kB" in line:
                    l = [float(x) for x in line.split()[2:]]
                    voigt = np.array([l[0], l[1], l[2], l[4], l[5], l[3]])
                    stress = ut.voigt2cartesian(voigt) * ureg("kbar")
                    warnings.warn(
                        "According to VASP this is kB units, but when comparing to QE it appears to be GPa.",
                        UserWarning,
                    )
        else:
            raise NotImplementedError("Unsupported filetype")
        lines.close()
    if "stress" not in locals():
        raise NameError("Stress tensor not found.")
    return stress


def kpath(file: str, labels: bool = True) -> SimpleNamespace | np.ndarray:
    """
    Greps the coordinates, labels and number of poiints from the path in reciprocal space.

    Currently supports:
    - QuantumEspresso: qe_bands_in, matdyn_in.
    - VASP: KPATH (KPOINTS in line mode).

    The code expects the labels to be after the high-symmetry points commented with a `!` as:
    ...
    0   0   0   ! Gamma
    0   0.5 0   ! X
    ...

    Parameters
    ----------
    file : str
        File from which to extract the kpath.
    labels : bool, optional
        Whether labels for the high-symmetry points are extracted. Default is True.

    Returns
    -------
    kpath : SimpleNamespace | np.ndarray
        If labels is True, a namespace with attributes `path` and `labels` is returned.
        Otherwise, the kpath is returned as an ndarray.


    Raises
    ------
    NameError:
        If label or kpath is not found.
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    """

    def read_qe_path(line_iter):
        kpath, k_names, N = [], [], None
        for line in line_iter:
            if N is None:
                N = int(line.split()[0])
            else:
                if labels:
                    try:
                        kpoint, label = line.split("!")
                    except ValueError:
                        raise NameError("Label not found, try using labels=False.")
                else:
                    kpoint = line
                # Grep K point
                kpoint = [float(x) for x in kpoint.split()]
                kpath.append(kpoint)
                # Grep K point label
                if labels:
                    new_name = label.split()[0]
                    k_names.append(new_name)
                # Check if path is complete
                if len(kpath) == N:
                    break
        return np.array(kpath), k_names

    filetype = _filetype(file)
    READ = EVEN = False
    kpath = k_names = N = None

    with open(file, "r") as lines:
        # QE input format
        if filetype in ["qe_bands_in", "matdyn_in"]:
            line_iter = iter(lines)
            for line in line_iter:
                if re.search("K_POINTS.*crystal_b", line, flags=re.IGNORECASE) or (
                    filetype == "matdyn_in" and re.search("/", line.split()[0])
                ):
                    kpath, k_names = read_qe_path(line_iter)
                    break
        # VASP KPATH format
        elif filetype == "kpath":
            for line in lines:
                # Grep number of points for each subpath
                if N is None:
                    try:
                        N = int(line.split()[0])
                    except ValueError:
                        pass
                elif re.search("Reciprocal", line, flags=re.IGNORECASE):
                    READ = True
                # Read path and labels
                elif READ:
                    if labels:
                        try:
                            kpoint, label = line.split("!")
                        except ValueError:
                            raise NameError("Label not found, try using labels=False.")
                    else:
                        kpoint = line
                    kpoint = [float(x) for x in kpoint.split()]
                    if kpath is None:
                        kpath = np.array([kpoint + [N]])
                        if labels:
                            k_names = [label.split()[0]]
                    # If points is different from previous in the KPATH file
                    elif (kpoint[:3] != kpath[-1][:3]).any():
                        if EVEN:
                            kpath = np.vstack([kpath, kpoint + [1]])
                        else:
                            kpath = np.vstack([kpath, kpoint + [N]])
                        if labels:
                            k_names = k_names + [label.split()[0]]
                    # If point is repeated and odd
                    elif not EVEN:
                        kpath[-1, -1] = N
                    EVEN = EVEN is False
        else:
            raise NotImplementedError("Unsupported filetype")
    if kpath is None:
        raise NameError("Kpath not found.")
    kpath = kpath * ureg._2pi / ureg.crystal
    if labels:
        # Post-process labels
        [l.replace("Gamma", r"\Gamma") for l in k_names]
        return SimpleNamespace(path=kpath, labels=k_names)
    else:
        return kpath


def kpointsEnergies(file: str) -> SimpleNamespace:
    """
    Grep the kpoints, energies, kpoint-weights  and orbital projections
    (if available) for different file kinds.

    Energies are given in eV and kpoints in reciprocal crystal units.
    Currently supports:
    - QuantumEspresso: qe_scf_out, projwfc_out, `.xml` files.
    - VASP: OUTCAR, EIGENVAL, PROCAR.

    Parameters
    ----------
    file : str
        File from which to extract the quantities.

    Returns
    -------
    SimpleNamespace : SimpleNamespace
        SimpleNamespace class with the following attributes:
        - energies : np.ndarray
            List of energies, each row corresponds to a particular k-point.
        - kpoints : np.ndarray
            List of k-points.
        - weights : np.ndarray
            List of kpoint-weights.
        - projections : _OrbitalProjectionContainer
            Container for orbital-resolved projection matrices.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    IndexError:
        If the number of weights does not coincide with the number of bands.
    """

    filetype = _filetype(file)
    READ_energies = READ_kpoints = RELAX_calc = RELAXED = OCCUPATIONS = False
    KPOINTS, ENERGIES, WEIGHTS, E, PROJ = [], [], [], [], []
    PROJECTIONS = None
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            return _Qe_xml(file).kpointsEnergies()
        elif filetype == "qe_scf_out":
            for line in lines:
                # Grep number of bands
                if "number of Kohn-Sham" in line:
                    num_bands = int(line.split("=")[1])
                elif "number of k points" in line:
                    num_points = int(line.split("=")[1].split()[0])
                elif " cart. coord." in line:
                    READ_kpoints = True
                elif "force convergence" in line:
                    RELAX_calc = True
                elif "Final scf calculation at the relaxed" in line:
                    RELAXED = True
                elif re.search("End of .* calculation", line):
                    if (RELAX_calc == False) or (
                        RELAX_calc == True and RELAXED == True
                    ):
                        READ_energies = True
                elif READ_kpoints:
                    k = [float(x) for x in line.split("(")[2].split(")")[0].split()]
                    w = float(line.split()[-1])
                    KPOINTS.append(k)
                    WEIGHTS.append(w)
                    if len(WEIGHTS) == num_points:
                        READ_kpoints = False
                elif READ_energies:
                    if line.lstrip().startswith("k ="):
                        OCCUPATIONS = False
                    elif OCCUPATIONS:
                        # Currently not implemented
                        pass
                    elif line.strip() != "":
                        for e in re.findall(r"[-+]?\d*\.\d+|\d+", line):
                            E.append(float(e))
                        if len(E) == num_bands:
                            ENERGIES.append(E)
                            E = []
                            OCCUPATIONS = True
            # Recover crystal units
            lat = lattice(file)
            lat = lat / np.linalg.norm(lat[0])
            Klat = ut.reciprocal_basis(lat).magnitude
            KPOINTS = ut.cartesian2cryst(KPOINTS, Klat) * (ureg._2pi / ureg.crystal)
            ENERGIES *= ureg("eV")
        elif filetype == "qe_proj_out":
            STATES, PROJECTIONS = [], []
            for line in lines:
                if "natomwfc" in line:
                    num_states = int(line.split()[-1])
                if "nbnd" in line:
                    num_bands = int(line.split()[-1])
                if "nkstot" in line:
                    num_points = int(line.split()[-1])
                # Scrape STATES info
                if "state #" in line:
                    l = line.split()
                    if "j=" in line:
                        # atom, l, j, mj, wfc
                        state = [
                            int(l[4]) - 1,
                            int(l[9][3:]),
                            float(l[10][2:]),
                            float(l[-1].split("=")[-1][:-1]),
                            int(l[8]) - 1,
                        ]
                        STATES.append(state)
                        soc_mode = True
                    if "m=" in line:
                        # atom, l, m, wfc
                        state = [
                            int(l[4]) - 1,
                            int(l[9][3:]),
                            int(l[-1][:-1]),
                            int(l[8]) - 1,
                        ]
                        STATES.append(state)
                        soc_mode = False
                if "k =" in line:
                    k = [float(x) for x in line.split()[2:]]
                    KPOINTS.append(k)
                elif "eV ===" in line:
                    energy = float(line.split()[-3])
                    E.append(energy)
                    P = np.zeros(num_states)
                    if len(E) == num_bands:
                        ENERGIES.append(E)
                        E = []
                elif "*[#" in line:
                    l = line.split("=")[-1].split("]")[:-1]
                    for x in l:
                        split = x.split("*[#")
                        try:
                            proj = float(split[0])
                        except ValueError:
                            print(f"ValueError with {split[0]} at {k}, defaulted to 0.")
                            proj = 0
                        index = int(split[1]) - 1
                        P[index] = proj
                if "|psi|" in line:
                    PROJ.append(P)
                    if len(PROJ) == num_bands:
                        PROJECTIONS.append(PROJ)
                        PROJ = []
            KPOINTS = np.array(KPOINTS) * (ureg._2pi / ureg.alat)
            ENERGIES *= ureg("eV")
            # Save projections into the proper container
            PROJ = np.array(PROJECTIONS)
            STATES = np.array(STATES)
            PROJECTIONS = _OrbitalProjectionContainer(code="quantum_espresso")
            N = 0
            if soc_mode:
                for i, S in enumerate(STATES):
                    PROJECTIONS.add_qe(
                        ion=S[0],
                        l=S[1],
                        wfc=S[4],
                        matrix=PROJ[:, :, i],
                        j=S[2],
                        mj=S[3],
                    )
            else:
                for i, S in enumerate(STATES):
                    PROJECTIONS.add_qe(
                        ion=S[0], l=S[1], wfc=S[3], matrix=PROJ[:, :, i], m=S[2]
                    )
        elif filetype == "eigenval":
            for i, line in enumerate(lines):
                l = line.split()
                if i == 5:
                    num_points, num_bands = int(l[1]), int(l[2])
                    READ_kpoints = READ_energies = True
                elif READ_kpoints:
                    # Kpoint line
                    if len(l) == 4:
                        k = [float(x) for x in l[:3]]
                        w = float(l[-1])
                        KPOINTS.append(k)
                        WEIGHTS.append(w)
                    # Energy line
                    elif len(l) == 3:
                        E.append(float(l[1]))
                        if len(E) == num_bands:
                            ENERGIES.append(E)
                            E = []
            KPOINTS *= ureg._2pi / ureg.crystal
            ENERGIES *= ureg("eV")
        elif filetype == "outcar":
            for line in lines:
                if "NBANDS" in line:
                    num_bands = int(line.split()[-1])
                elif "Coordinates" in line and len(KPOINTS) == 0:
                    READ_kpoints = True
                elif "band No." in line:
                    READ_energies = True
                elif READ_kpoints:
                    l = line.split()
                    if len(l) != 0:
                        k = [float(x) for x in l[:3]]
                        w = float(l[-1])
                        KPOINTS.append(k)
                        WEIGHTS.append(w)
                    else:
                        num_points = len(KPOINTS)
                        READ_kpoints = False
                elif READ_energies:
                    l = line.split()
                    if len(l) == 3:
                        E.append(float(l[1]))
                        if len(E) == num_bands:
                            ENERGIES.append(E)
                            E = []
                            if len(ENERGIES) == num_points:
                                break
            KPOINTS *= ureg._2pi / ureg.crystal
            ENERGIES *= ureg("eV")
        elif filetype == "procar":
            for line in lines:
                if "k-points" in line:
                    l = line.split()
                    num_points = int(l[3])
                    num_bands = int(l[7])
                    num_ions = int(l[-1])
                elif "k-point" in line:
                    numbers = [float(x) for x in re.findall(r"[-+]?\d*\.\d+|\d+", line)]
                    k = numbers[1:4]
                    w = numbers[-1]
                    KPOINTS.append(k)
                    WEIGHTS.append(w)
                elif "energy" in line:
                    E.append(float(line.split()[4]))
                    if len(E) == num_bands:
                        ENERGIES.append(E)
                        E = []
                elif "tot" not in line and len(line.split()) == 11:
                    proj = [float(x) for x in line.split()[1:-1]]
                    PROJ.append(proj)
            KPOINTS *= ureg._2pi / ureg.crystal
            ENERGIES *= ureg("eV")
            # Save projections into the proper container
            PROJ = np.array(PROJ)
            M = round(PROJ.shape[0] / (np.prod(np.shape(ENERGIES)) * num_ions))
            PROJECTIONS = _OrbitalProjectionContainer(code="vasp")
            for i in range(num_ions):
                for l in range(3):
                    for m in range(-l, l + 1):
                        for mag in range(M):
                            C = l + l * l + m
                            matrix = PROJ[
                                i + num_ions * mag :: num_ions * M, C
                            ].reshape(np.shape(ENERGIES))
                            PROJECTIONS.add_poscar(i, l, m, mag, matrix)
        else:
            raise NotImplementedError("Unsupported filetype")
    WEIGHTS = np.array(WEIGHTS)
    if WEIGHTS.shape[0] != ENERGIES.shape[0] and len(WEIGHTS) != 0:
        raise IndexError(
            f"There should be as many `weights` ({WEIGHTS.shape[0]}) as bands ({ENERGIES.shape[0]})."
        )
    return SimpleNamespace(
        energies=ENERGIES,
        kpoints=KPOINTS,
        weights=WEIGHTS,
        projections=PROJECTIONS,
    )


def kpointsFrequencies(file: str) -> SimpleNamespace:
    """
    Grep the kpoints and frequencies from phonon ouputs.

    Frequencies are given in cm-1 and kpoints in reciprocal alat.
    Currently supports:
    - QuantumEspresso: qe_freq_out.

    Parameters
    ----------
    file : str
        File from which to extract the spectrum.

    Returns
    -------
    SimpleNamespace : SimpleNamespace
        SimpleNamespace class with the following attributes:
        - frequencies : np.ndarray
            List of frequencies, each row corresponds to a particular k-point.
        - kpoints : np.ndarray
            List of k-points.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    """

    filetype = _filetype(file)
    KPOINTS, FREQS, F = [], [], []
    READ_freqs = False
    with open(file, "r") as lines:
        if filetype == "qe_freq_out":
            for line in lines:
                l = line.split()
                if "nbnd" in line:
                    num_bands = int(l[2][:-1])
                    num_points = int(line.split("nks=")[-1][:-2])
                elif len(l) == 3:
                    k = [float(x) for x in l]
                    KPOINTS.append(k)
                    READ_freqs = True
                elif READ_freqs:
                    for f in l:
                        F.append(float(f))
                    if len(F) == num_bands:
                        FREQS.append(F)
                        F = []
        else:
            raise NotImplementedError("Unsupported filetype")
    # Give proper units
    FREQS = np.array(FREQS) * ureg("c") / ureg("cm")
    KPOINTS = np.array(KPOINTS) * (ureg("_2pi") / ureg("alat"))
    return SimpleNamespace(frequencies=FREQS, kpoints=KPOINTS)


def dyn_file(file: str) -> SimpleNamespace:
    """
    Parse a dynamical matrix file and extract phonon mode information.

    This function extracts:
    - the lattice vectors,
    - the atomic species and their masses,
    - the atom types and positions,
    - the q-point at which the phonon modes are computed,
    - the phonon frequencies (in cm⁻¹),
    - and the polarization vectors (phonon eigenvectors).

    Parameters
    ----------
    file : str
        Path to the dynamical matrix file (e.g. QE `.dyn` or `.dynmat` file).

    Returns
    -------
    SimpleNamespace
        A container with the following fields:
        - q : ureg.Quantity, shape (3,)
            The q-point in where the calculation was performed.
        - lattice : ureg.Quantity, shape (3, 3)
            The lattice vectors of the unit cell.
        - freqs : ureg.Quantity, shape (n_modes,)
            Array of vibrational frequencies (in cm⁻¹).
        - displacements : np.ndarray, shape(n_modes,n_atoms,3)
            An (n_modes, n_atoms, 3) array of complex normalized displacement vectors for each mode.
        - positions : ureg.Quantity, shape (n_atoms, 3)
            Atomic positions.
        - elements : list of str
            Chemical symbol for each atom.
        - masses : ureg.Quantity, shape (n_atoms,)
            Atomic mass in atomic units for each atom.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    """
    filetype = _filetype(file)
    if filetype != "qe_dyn":
        raise NotImplementedError("Unsupported filetype")
    lat = lattice(file)
    n_atoms = n_types = freqs = vec = alat = None
    vec, freqs = [], []
    species = []
    atoms = []
    displacements = []
    read_modes = False
    with open(file, "r") as lines:
        for line in lines:
            l = line.split()
            if n_atoms is None and len(l) == 9:
                # Get number of species and atoms
                l = line.split()
                n_types, n_atoms = int(l[0]), int(l[1])
                alat = float(l[3]) * ureg("bohr")
            elif n_types != 0 and len(l) == 4:
                # Get species
                new = [int(l[0]), l[1][1:], float(l[-1])]
                species.append(new)
                n_types -= 1
            elif n_atoms != 0 and len(l) == 5:
                # Get atomic positions
                new = [int(l[1])] + [float(x) for x in l[2:]]
                atoms.append(new)
                n_atoms -= 1
            elif "Diagonalizing" in line:
                read_modes = True
            elif read_modes:
                # Read modes
                if "q = (" in line:
                    q_point = np.array([float(x) for x in l[3:6]])
                elif "freq" in line:
                    freqs.append(float(l[-2]))
                elif len(l) == 8:
                    nums = l[1:-1]
                    new = [
                        complex(float(nums[i]), float(nums[i + 1]))
                        for i in range(0, 6, 2)
                    ]
                    vec.append(new)
                    if len(vec) == len(atoms):
                        displacements.append(vec)
                        vec = []

    # Attach units and get positions, elmenets and masses.
    positions = (np.array(atoms)[:, 1:] * alat).to("ang")
    indices = np.array(atoms)[:, 0].astype(int) - 1
    elements = [species[x][1] for x in indices]
    masses = np.array([species[x][2] for x in indices]) * ureg._2m_e
    displacements = np.array(displacements)
    q_point = (q_point * ureg._2pi / alat).to("_2pi/ang")
    freqs = np.array(freqs) * ureg("c/cm")

    return SimpleNamespace(
        q=q_point,
        lattice=lat,
        freqs=freqs,
        displacements=displacements,
        positions=positions,
        elements=elements,
        masses=masses,
    )


def _find_dyn_file(q_cryst: np.ndarray | ureg.Quantity, results_ph_path: str) -> str:
    """
    Search for the Quantum ESPRESSO `.dyn` file containing a specified q-point
    in crystalline coordinates.

    This function compares the requested q-point with those found in each
    `dyn*` file, accounting for equivalence under reciprocal lattice translations.

    Parameters
    ----------
    q_cryst : np.ndarray | ureg.Quantity
        The q-point to locate, expressed in crystalline (reduced) coordinates.
    results_ph_path : str
        Path to the folder where the phonon (`ph.x`) output `.dyn*` files are stored.

    Returns
    -------
    str
        The full path to the `.dyn` file that contains the matching q-point.

    Raises
    ------
    FileNotFoundError
        If no `.dyn*` file or no matching q-point is found in any of the `.dyn` files
    """
    from glob import glob

    # Locate a reference .dyn file to extract lattice
    dyn1 = glob(results_ph_path + "/*dyn1") + glob(results_ph_path + "/*dyn")
    if not dyn1:
        raise FileNotFoundError(
            "No 'dyn1' or 'dyn' file found in the specified folder."
        )

    # Read lattice and convert to alat units
    lattice = dyn_file(dyn1[0]).lattice
    lattice = lattice / np.linalg.norm(lattice[0]) * ureg.alat
    k_basis = ut.reciprocal_basis(lattice)

    # Scan all .dyn* files (excluding matdyn if present)
    dyn_files = glob(results_ph_path + "/*.dyn*")
    dyn_files = [f for f in dyn_files if "results_matdyn" not in f]

    for file in dyn_files:
        with open(file, "r") as f:
            for line in f:
                if "q = (" in line:
                    q_point = np.array([float(x) for x in line.split()[3:6]]) * ureg(
                        "_2pi/alat"
                    )
                    q_crys_from_file = ut.cartesian2cryst(q_point, k_basis)

                    # Account for periodic images using symmetry
                    for q_equiv in ut._expand_zone_border(q_crys_from_file):
                        if np.allclose(q_cryst, q_equiv, atol=1e-4):
                            return file

    raise FileNotFoundError(
        f"No `.dyn*` file found containing q = {q_cryst} in crystalline coordinates."
    )


def dyn_q(
    q_cryst: np.ndarray | ureg.Quantity, results_ph_path: str, qe_format: bool = True
) -> SimpleNamespace:
    """
    Reads the Quantum ESPRESSO `.dyn*` file corresponding to a given q-point.

    This function locates the `.dyn*` file generated by `ph.x` that corresponds to a
    desired q-point (in reduced crystalline coordinates), extracts the corresponding dynamical
    matrix, and optionally converts it to the real physical dynamical matrix in units
    of 1 / [time]².

    Parameters
    ----------
    q_cryst : np.ndarray | ureg.Quantity
        The q-point of interest, expressed in reduced crystalline coordinates (fractions of reciprocal lattice vectors).
        If not a `Quantity`, it is assumed to be in `_2pi/crystal` units.

    results_ph_path : str
        Path to the directory containing the Quantum ESPRESSO `ph.x` output `.dyn*` files.

    qe_format : bool, optional
        If True (default), returns the raw QE dynamical matrix (includes sqrt(m_i m_j) mass factors).
        If False, converts the dynamical matrix to true physical form in 1 / [time]² units (Ry/h)^2.

    Returns
    -------
    system : SimpleNamespace
        A container with the following fields:
        - q : ureg.Quantity, shape (3,)
            The q-point in where the calculation was performed.
        - lattice : ureg.Quantity, shape (3, 3)
            The lattice vectors of the unit cell.
        - freqs : ureg.Quantity, shape (n_modes,)
            Array of vibrational frequencies (in cm⁻¹).
        - positions : ureg.Quantity, shape (n_atoms, 3)
            Atomic positions.
        - elements : list of str
            Chemical symbol for each atom.
        - masses : ureg.Quantity, shape (n_atoms,)
            Atomic mass in atomic units for each atom.
        - dyn: ureg.Quantity
            The (3N × 3N) complex dynamical matrix (units depend on `qe_format`).

    Notes
    -----
    - The dynamical matrix read from QE includes a sqrt(m_i m_j) prefactor for each (3×3) subblock.
      This must be removed to obtain the physical matrix for diagonalization (ω² in Ry²/ħ²).
    - The units of the returned matrix are `_2m_e * Ry^2 / planck_constant^2` in QE format.
    """
    from yaiv.phonon import _QEdyn2Realdyn

    # Normalize units
    if isinstance(q_cryst, ureg.Quantity):
        q_cryst = q_cryst.to("_2pi/crystal")
    else:
        q_cryst = q_cryst * ureg("_2pi/crystal")

    file = _find_dyn_file(q_cryst, results_ph_path)

    system = dyn_file(file)
    dim = 3 * len(system.elements)
    dyn_mat = np.zeros((dim, dim), dtype=complex)

    # Lattice and reciprocal basis
    lattice = system.lattice / np.linalg.norm(system.lattice[0]) * ureg.alat
    k_basis = ut.reciprocal_basis(lattice)

    READ_dynmat = False
    with open(file, "r") as lines:
        for line in lines:
            if "q = (" in line:
                q_point = np.array([float(x) for x in line.split()[3:6]]) * ureg(
                    "_2pi/alat"
                )
                q_crys_from_file = ut.cartesian2cryst(q_point, k_basis)

                for q_equiv in ut._expand_zone_border(q_crys_from_file):
                    if np.allclose(q_cryst.magnitude, q_equiv.magnitude, atol=1e-4):
                        READ_dynmat = True
                        break

            if READ_dynmat:
                l = line.split()
                if len(l) == 2:  # matrix block index line
                    n, m = int(l[0]), int(l[1])
                    num = 0
                elif len(l) == 6:  # submatrix row
                    row = np.array(
                        [
                            complex(float(l[0]), float(l[1])),
                            complex(float(l[2]), float(l[3])),
                            complex(float(l[4]), float(l[5])),
                        ]
                    )
                    sub_mat = row if num == 0 else np.vstack([sub_mat, row])
                    num += 1
                    if num == 3:
                        i, j = 3 * (n - 1), 3 * (m - 1)
                        dyn_mat[i : i + 3, j : j + 3] = sub_mat
                if re.search("Dynamical", line) or re.search("Diagonalizing", line):
                    break

    # Clean output
    system.q = q_cryst
    delattr(system, "displacements")
    system.dyn = dyn_mat * ureg("_2m_e * Ry^2 / planck_constant^2")
    if not qe_format:
        system.dyn = _QEdyn2Realdyn(system.dyn, system.masses)

    return system


def symmetries(file: str) -> list[SimpleNamespace]:
    """
    Grep symmetry operations and return them as rotation/translation pairs.

    This reads all symmetry elements, extracts the 3×3 rotation matrix and
    a fractional translation:
      - R: np.ndarray shape (3, 3) in crystal coord.
      - t: ureg.Quantity length-3 vector in units of 2π/crystal

    Parameters
    ----------
    file : str
        Path to the output file containing symmetry information (e.g., QE XML).

    Returns
    -------
    symmetries : list[SimpleNamespace]
        Each entry contains R (3×3 rotation matrix) and t (fractional translation
        vector in 2π/crystal units).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    """
    filetype = _filetype(file)
    if filetype == "qe_xml":
        symmetries = _Qe_xml(file).symmetries()
    else:
        raise NotImplementedError("Unsupported filetype")
    return symmetries


def cutoff(file: str) -> ureg.Quantity:
    """
    Greps the cutoff energy from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the cutoff energy.

    Returns
    -------
    cutoff : ureg.Quantity
        Cutoff energy with attached units (ureg.Quantity).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The cutoff energy was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            cutoff = _Qe_xml(file).cutoff()
        elif filetype == "qe_scf_out":
            for line in lines:
                # If smearing is used
                if "kinetic-energy cutoff" in line:
                    cutoff = float(line.split()[-2])
                    break
            cutoff *= ureg("Ry")
        else:
            raise NotImplementedError("Unsupported filetype")
    if "cutoff" not in locals():
        raise NameError("Cutoff energy not found.")
    return cutoff


def smearing(file: str) -> ureg.Quantity:
    """
    Greps the smearing from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the smearing.

    Returns
    -------
    smearing : ureg.Quantity
        Smearing with attached units (ureg.Quantity).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The smearing was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            smearing = _Qe_xml(file).smearing()
        elif filetype == "qe_scf_out":
            for line in lines:
                if "smearing, width" in line:
                    smearing = float(line.split()[-1])
                    break
            smearing *= ureg("Ry")
        else:
            raise NotImplementedError("Unsupported filetype")
    if "smearing" not in locals():
        raise NameError("Smearing energy not found.")
    return smearing


def k_grid(file: str) -> list[int]:
    """
    Greps the k-grid from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the k-grid.

    Returns
    -------
    k_grid : list(int)
        K-grid used in the computation.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The k-grid was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            kgrid = _Qe_xml(file).k_grid()
        elif filetype == "qe_scf_in":
            READ = False
            for line in lines:
                if "K_POINTS" in line:
                    READ = True
                elif READ:
                    l = line.split()
                    if len(l) >= 3:
                        kgrid = [int(x) for x in l[:3]]
                        break
        elif filetype == "qe_scf_out":
            # Try to find the corresponding input file
            kgrid = k_grid(file[:-1] + "i")
        else:
            raise NotImplementedError("Unsupported filetype")
    if "kgrid" not in locals():
        raise NameError("K-grid not found.")
    return kgrid


def atomic_forces(file: str) -> SimpleNamespace:
    """
    Greps the atomic forces from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the atomic forces.

    Returns
    -------
    forces : SimpleNamespace
        SimpleNamespace class with the following attributes:
        SimpleNamespace with atomic forces with attached units (ureg.Quantity).
        - per_atom : np.ndarray | ureg.Quantity
            List of foces per atom with shape (N,3) where N is the number of atoms.
        - total : np.ndarray | ureg.Quantity
            Total average force.

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The atomic forces were not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_scf_out":
            READ, WRITE = False, False
            forces = []
            for line in lines:
                if "Forces acting on atoms" in line:
                    READ, WRITE = True, True
                elif READ:
                    l = line.split()
                    if "atom" in line and WRITE:
                        f = [float(x) for x in l[6 : 6 + 3]]
                        forces.append(f)
                    elif "Total force" in line:
                        total_force = float(l[3])
                        break
                    elif len(forces) != 0:
                        WRITE = False
            forces = np.asarray(forces) * ureg("Ry/bohr")
            total_force *= ureg("Ry/bohr")
        else:
            raise NotImplementedError("Unsupported filetype")
    if "forces" not in locals():
        raise NameError("Atomic forces not found.")
    return SimpleNamespace(per_atom=forces, total=total_force)


def runtime(file: str) -> ureg.Quantity:
    """
    Greps the computational runtime from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the computational runtime.

    Returns
    -------
    runtime : ureg.Quantity
        Computational runtime with attached units (ureg.Quantity).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The computational runtime was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_xml":
            runtime = _Qe_xml(file).runtime()
        elif filetype == "qe_scf_out":
            for line in reversed(list(lines)):
                # If smearing is used
                if "PWSCF" in line:
                    h, m, s = 0, 0, 0
                    runtime = line.split()[2]
                    if "h" in runtime:
                        h = int(runtime.split("h")[0])
                        runtime = runtime.split("h")[1]
                    if "m" in runtime:
                        m = int(runtime.split("m")[0])
                        runtime = runtime.split("m")[1]
                    if "s" in runtime:
                        s = float(runtime.split("s")[0])
                    runtime = s * ureg.second + m * ureg.minute + h * ureg.hour
                    break
        else:
            raise NotImplementedError("Unsupported filetype")
    if "runtime" not in locals():
        raise NameError("Computation runtime not found.")
    return runtime


def ram(file: str) -> ureg.Quantity:
    """
    Greps the RAM needed in the computation from a variety of filetypes.

    Parameters
    ----------
    file : str
        File from which to extract the RAM.

    Returns
    -------
    RAM : ureg.Quantity
        RAM with attached units (ureg.Quantity).

    Raises
    ------
    NotImplementedError:
        The function is not currently implemeted for the provided filetype.
    NameError:
        The RAM was not found.
    """
    filetype = _filetype(file)
    with open(file, "r") as lines:
        if filetype == "qe_scf_out":
            for line in lines:
                if "total dynamical RAM" in line:
                    RAM = float(line.split()[5])
                    units = ureg(line.split()[6])
                    break
            RAM *= units
        else:
            raise NotImplementedError("Unsupported filetype")
    if "RAM" not in locals():
        raise NameError("RAM not found.")
    return RAM
